# This file is part of cloud-init. See LICENSE file for license information.

import re

from cloudinit import distros, ssh_util
from tests.unittests.helpers import CiTestCase, mock
from tests.unittests.util import abstract_to_concrete


@mock.patch("cloudinit.distros.util.system_is_snappy", return_value=False)
@mock.patch("cloudinit.distros.subp.subp")
class TestCreateUser(CiTestCase):

    with_logs = True

    def setUp(self):
        super(TestCreateUser, self).setUp()
        self.dist = abstract_to_concrete(distros.Distro)(
            name="test", cfg=None, paths=None
        )

    def _useradd2call(self, args):
        # return a mock call for the useradd command in args
        # with expected 'logstring'.
        args = ["useradd"] + args
        logcmd = [a for a in args]
        for i in range(len(args)):
            if args[i] in ("--password",):
                logcmd[i + 1] = "REDACTED"
        return mock.call(args, logstring=logcmd)

    def test_basic(self, m_subp, m_is_snappy):
        user = "foouser"
        self.dist.create_user(user)
        self.assertEqual(
            m_subp.call_args_list,
            [
                self._useradd2call([user, "-m"]),
                mock.call(["passwd", "-l", user]),
            ],
        )

    def test_no_home(self, m_subp, m_is_snappy):
        user = "foouser"
        self.dist.create_user(user, no_create_home=True)
        self.assertEqual(
            m_subp.call_args_list,
            [
                self._useradd2call([user, "-M"]),
                mock.call(["passwd", "-l", user]),
            ],
        )

    def test_system_user(self, m_subp, m_is_snappy):
        # system user should have no home and get --system
        user = "foouser"
        self.dist.create_user(user, system=True)
        self.assertEqual(
            m_subp.call_args_list,
            [
                self._useradd2call([user, "--system", "-M"]),
                mock.call(["passwd", "-l", user]),
            ],
        )

    def test_explicit_no_home_false(self, m_subp, m_is_snappy):
        user = "foouser"
        self.dist.create_user(user, no_create_home=False)
        self.assertEqual(
            m_subp.call_args_list,
            [
                self._useradd2call([user, "-m"]),
                mock.call(["passwd", "-l", user]),
            ],
        )

    def test_unlocked(self, m_subp, m_is_snappy):
        user = "foouser"
        self.dist.create_user(user, lock_passwd=False)
        self.assertEqual(
            m_subp.call_args_list, [self._useradd2call([user, "-m"])]
        )

    def test_set_password(self, m_subp, m_is_snappy):
        user = "foouser"
        password = "passfoo"
        self.dist.create_user(user, passwd=password)
        self.assertEqual(
            m_subp.call_args_list,
            [
                self._useradd2call([user, "--password", password, "-m"]),
                mock.call(["passwd", "-l", user]),
            ],
        )

    @mock.patch("cloudinit.distros.util.is_group")
    def test_group_added(self, m_is_group, m_subp, m_is_snappy):
        m_is_group.return_value = False
        user = "foouser"
        self.dist.create_user(user, groups=["group1"])
        expected = [
            mock.call(["groupadd", "group1"]),
            self._useradd2call([user, "--groups", "group1", "-m"]),
            mock.call(["passwd", "-l", user]),
        ]
        self.assertEqual(m_subp.call_args_list, expected)

    @mock.patch("cloudinit.distros.util.is_group")
    def test_only_new_group_added(self, m_is_group, m_subp, m_is_snappy):
        ex_groups = ["existing_group"]
        groups = ["group1", ex_groups[0]]
        m_is_group.side_effect = lambda m: m in ex_groups
        user = "foouser"
        self.dist.create_user(user, groups=groups)
        expected = [
            mock.call(["groupadd", "group1"]),
            self._useradd2call([user, "--groups", ",".join(groups), "-m"]),
            mock.call(["passwd", "-l", user]),
        ]
        self.assertEqual(m_subp.call_args_list, expected)

    @mock.patch("cloudinit.distros.util.is_group")
    def test_create_groups_with_whitespace_string(
        self, m_is_group, m_subp, m_is_snappy
    ):
        # groups supported as a comma delimeted string even with white space
        m_is_group.return_value = False
        user = "foouser"
        self.dist.create_user(user, groups="group1, group2")
        expected = [
            mock.call(["groupadd", "group1"]),
            mock.call(["groupadd", "group2"]),
            self._useradd2call([user, "--groups", "group1,group2", "-m"]),
            mock.call(["passwd", "-l", user]),
        ]
        self.assertEqual(m_subp.call_args_list, expected)

    def test_explicit_sudo_false(self, m_subp, m_is_snappy):
        user = "foouser"
        self.dist.create_user(user, sudo=False)
        self.assertEqual(
            m_subp.call_args_list,
            [
                self._useradd2call([user, "-m"]),
                mock.call(["passwd", "-l", user]),
            ],
        )

    @mock.patch("cloudinit.ssh_util.setup_user_keys")
    def test_setup_ssh_authorized_keys_with_string(
        self, m_setup_user_keys, m_subp, m_is_snappy
    ):
        """ssh_authorized_keys allows string and calls setup_user_keys."""
        user = "foouser"
        self.dist.create_user(user, ssh_authorized_keys="mykey")
        self.assertEqual(
            m_subp.call_args_list,
            [
                self._useradd2call([user, "-m"]),
                mock.call(["passwd", "-l", user]),
            ],
        )
        m_setup_user_keys.assert_called_once_with(set(["mykey"]), user)

    @mock.patch("cloudinit.ssh_util.setup_user_keys")
    def test_setup_ssh_authorized_keys_with_list(
        self, m_setup_user_keys, m_subp, m_is_snappy
    ):
        """ssh_authorized_keys allows lists and calls setup_user_keys."""
        user = "foouser"
        self.dist.create_user(user, ssh_authorized_keys=["key1", "key2"])
        self.assertEqual(
            m_subp.call_args_list,
            [
                self._useradd2call([user, "-m"]),
                mock.call(["passwd", "-l", user]),
            ],
        )
        m_setup_user_keys.assert_called_once_with(set(["key1", "key2"]), user)

    @mock.patch("cloudinit.ssh_util.setup_user_keys")
    def test_setup_ssh_authorized_keys_with_integer(
        self, m_setup_user_keys, m_subp, m_is_snappy
    ):
        """ssh_authorized_keys warns on non-iterable/string type."""
        user = "foouser"
        self.dist.create_user(user, ssh_authorized_keys=-1)
        m_setup_user_keys.assert_called_once_with(set([]), user)
        match = re.match(
            r".*WARNING: Invalid type \'<(type|class) \'int\'>\' detected for"
            " 'ssh_authorized_keys'.*",
            self.logs.getvalue(),
            re.DOTALL,
        )
        self.assertIsNotNone(
            match, "Missing ssh_authorized_keys invalid type warning"
        )

    @mock.patch("cloudinit.ssh_util.setup_user_keys")
    def test_create_user_with_ssh_redirect_user_no_cloud_keys(
        self, m_setup_user_keys, m_subp, m_is_snappy
    ):
        """Log a warning when trying to redirect a user no cloud ssh keys."""
        user = "foouser"
        self.dist.create_user(user, ssh_redirect_user="someuser")
        self.assertIn(
            "WARNING: Unable to disable SSH logins for foouser given "
            "ssh_redirect_user: someuser. No cloud public-keys present.\n",
            self.logs.getvalue(),
        )
        m_setup_user_keys.assert_not_called()

    @mock.patch("cloudinit.ssh_util.setup_user_keys")
    def test_create_user_with_ssh_redirect_user_with_cloud_keys(
        self, m_setup_user_keys, m_subp, m_is_snappy
    ):
        """Disable ssh when ssh_redirect_user and cloud ssh keys are set."""
        user = "foouser"
        self.dist.create_user(
            user, ssh_redirect_user="someuser", cloud_public_ssh_keys=["key1"]
        )
        disable_prefix = ssh_util.DISABLE_USER_OPTS
        disable_prefix = disable_prefix.replace("$USER", "someuser")
        disable_prefix = disable_prefix.replace("$DISABLE_USER", user)
        m_setup_user_keys.assert_called_once_with(
            set(["key1"]), "foouser", options=disable_prefix
        )

    @mock.patch("cloudinit.ssh_util.setup_user_keys")
    def test_create_user_with_ssh_redirect_user_does_not_disable_auth_keys(
        self, m_setup_user_keys, m_subp, m_is_snappy
    ):
        """Do not disable ssh_authorized_keys when ssh_redirect_user is set."""
        user = "foouser"
        self.dist.create_user(
            user,
            ssh_authorized_keys="auth1",
            ssh_redirect_user="someuser",
            cloud_public_ssh_keys=["key1"],
        )
        disable_prefix = ssh_util.DISABLE_USER_OPTS
        disable_prefix = disable_prefix.replace("$USER", "someuser")
        disable_prefix = disable_prefix.replace("$DISABLE_USER", user)
        self.assertEqual(
            m_setup_user_keys.call_args_list,
            [
                mock.call(set(["auth1"]), user),  # not disabled
                mock.call(set(["key1"]), "foouser", options=disable_prefix),
            ],
        )

    @mock.patch("cloudinit.distros.subp.which")
    def test_lock_with_usermod_if_no_passwd(
        self, m_which, m_subp, m_is_snappy
    ):
        """Lock uses usermod --lock if no 'passwd' cmd available."""
        m_which.side_effect = lambda m: m in ("usermod",)
        self.dist.lock_passwd("bob")
        self.assertEqual(
            [mock.call(["usermod", "--lock", "bob"])], m_subp.call_args_list
        )

    @mock.patch("cloudinit.distros.subp.which")
    def test_lock_with_passwd_if_available(self, m_which, m_subp, m_is_snappy):
        """Lock with only passwd will use passwd."""
        m_which.side_effect = lambda m: m in ("passwd",)
        self.dist.lock_passwd("bob")
        self.assertEqual(
            [mock.call(["passwd", "-l", "bob"])], m_subp.call_args_list
        )

    @mock.patch("cloudinit.distros.subp.which")
    def test_lock_raises_runtime_if_no_commands(
        self, m_which, m_subp, m_is_snappy
    ):
        """Lock with no commands available raises RuntimeError."""
        m_which.return_value = None
        with self.assertRaises(RuntimeError):
            self.dist.lock_passwd("bob")


# vi: ts=4 expandtab



# This file is part of cloud-init. See LICENSE file for license information.


import importlib
import inspect
import itertools
import logging
import sys
from copy import copy
from pathlib import Path
from textwrap import dedent

import pytest
import yaml
from yaml import safe_load

from cloudinit.config.schema import (
    CLOUD_CONFIG_HEADER,
    MetaSchema,
    SchemaValidationError,
    _schemapath_for_cloudconfig,
    annotated_cloudconfig_file,
    get_jsonschema_validator,
    get_meta_doc,
    get_schema,
    load_doc,
    main,
    validate_cloudconfig_file,
    validate_cloudconfig_metaschema,
    validate_cloudconfig_schema,
)
from cloudinit.util import write_file
from tests.unittests.helpers import (
    CiTestCase,
    cloud_init_project_dir,
    mock,
    skipUnlessJsonSchema,
)


def get_schemas() -> dict:
    """Return all legacy module schemas

    Assumes that module schemas have the variable name "schema"
    """
    return get_module_variable("schema")


def get_metas() -> dict:
    """Return all module metas

    Assumes that module schemas have the variable name "schema"
    """
    return get_module_variable("meta")


def get_module_variable(var_name) -> dict:
    """Inspect modules and get variable from module matching var_name"""
    schemas = {}

    files = list(
        Path(cloud_init_project_dir("cloudinit/config/")).glob("cc_*.py")
    )

    modules = [mod.stem for mod in files]

    for module in modules:
        importlib.import_module("cloudinit.config.{}".format(module))

    for k, v in sys.modules.items():
        path = Path(k)

        if "cloudinit.config" == path.stem and path.suffix[1:4] == "cc_":
            module_name = path.suffix[1:]
            members = inspect.getmembers(v)
            schemas[module_name] = None
            for name, value in members:
                if name == var_name:
                    schemas[module_name] = value
                    break
    return schemas


class TestGetSchema:
    def test_get_schema_coalesces_known_schema(self):
        """Every cloudconfig module with schema is listed in allOf keyword."""
        schema = get_schema()
        assert sorted(
            [
                "cc_apk_configure",
                "cc_apt_configure",
                "cc_apt_pipelining",
                "cc_bootcmd",
                "cc_byobu",
                "cc_ca_certs",
                "cc_chef",
                "cc_debug",
                "cc_disable_ec2_metadata",
                "cc_disk_setup",
                "cc_install_hotplug",
                "cc_keyboard",
                "cc_locale",
                "cc_ntp",
                "cc_resizefs",
                "cc_runcmd",
                "cc_snap",
                "cc_ubuntu_advantage",
                "cc_ubuntu_drivers",
                "cc_write_files",
                "cc_zypper_add_repo",
            ]
        ) == sorted(
            [meta["id"] for meta in get_metas().values() if meta is not None]
        )
        assert "http://json-schema.org/draft-04/schema#" == schema["$schema"]
        assert ["$defs", "$schema", "allOf"] == sorted(list(schema.keys()))
        # New style schema should be defined in static schema file in $defs
        expected_subschema_defs = [
            {"$ref": "#/$defs/cc_apk_configure"},
            {"$ref": "#/$defs/cc_apt_configure"},
            {"$ref": "#/$defs/cc_apt_pipelining"},
            {"$ref": "#/$defs/cc_bootcmd"},
            {"$ref": "#/$defs/cc_byobu"},
            {"$ref": "#/$defs/cc_ca_certs"},
            {"$ref": "#/$defs/cc_chef"},
            {"$ref": "#/$defs/cc_debug"},
            {"$ref": "#/$defs/cc_disable_ec2_metadata"},
            {"$ref": "#/$defs/cc_disk_setup"},
        ]
        found_subschema_defs = []
        legacy_schema_keys = []
        for subschema in schema["allOf"]:
            if "$ref" in subschema:
                found_subschema_defs.append(subschema)
            else:  # Legacy subschema sourced from cc_* module 'schema' attr
                legacy_schema_keys.extend(subschema["properties"].keys())

        assert expected_subschema_defs == found_subschema_defs
        # This list will dwindle as we move legacy schema to new $defs
        assert [
            "drivers",
            "keyboard",
            "locale",
            "locale_configfile",
            "ntp",
            "resize_rootfs",
            "runcmd",
            "snap",
            "ubuntu_advantage",
            "updates",
            "write_files",
            "write_files",
            "zypper",
        ] == sorted(legacy_schema_keys)


class TestLoadDoc:

    docs = get_module_variable("__doc__")

    # TODO( Drop legacy test when all sub-schemas in cloud-init-schema.json )
    @pytest.mark.parametrize(
        "module_name",
        (
            "cc_apt_pipelining",  # new style composite schema file
            "cc_zypper_add_repo",  # legacy sub-schema defined in module
        ),
    )
    def test_report_docs_for_legacy_and_consolidated_schema(self, module_name):
        doc = load_doc([module_name])
        assert doc, "Unexpected empty docs for {}".format(module_name)
        assert self.docs[module_name] == doc


class Test_SchemapathForCloudconfig:
    """Coverage tests for supported YAML formats."""

    @pytest.mark.parametrize(
        "source_content, expected",
        (
            (b"{}", {}),  # assert empty config handled
            # Multiple keys account for comments and whitespace lines
            (b"#\na: va\n  \nb: vb\n#\nc: vc", {"a": 2, "b": 4, "c": 6}),
            # List items represented on correct line number
            (b"a:\n - a1\n\n - a2\n", {"a": 1, "a.0": 2, "a.1": 4}),
            # Nested dicts represented on correct line number
            (b"a:\n a1:\n\n  aa1: aa1v\n", {"a": 1, "a.a1": 2, "a.a1.aa1": 4}),
        ),
    )
    def test_schemapaths_representatative_of_source_yaml(
        self, source_content, expected
    ):
        """Validate schemapaths dict accurately represents source YAML line."""
        cfg = yaml.safe_load(source_content)
        assert expected == _schemapath_for_cloudconfig(
            config=cfg, original_content=source_content
        )


class SchemaValidationErrorTest(CiTestCase):
    """Test validate_cloudconfig_schema"""

    def test_schema_validation_error_expects_schema_errors(self):
        """SchemaValidationError is initialized from schema_errors."""
        errors = (
            ("key.path", 'unexpected key "junk"'),
            ("key2.path", '"-123" is not a valid "hostname" format'),
        )
        exception = SchemaValidationError(schema_errors=errors)
        self.assertIsInstance(exception, Exception)
        self.assertEqual(exception.schema_errors, errors)
        self.assertEqual(
            'Cloud config schema errors: key.path: unexpected key "junk", '
            'key2.path: "-123" is not a valid "hostname" format',
            str(exception),
        )
        self.assertTrue(isinstance(exception, ValueError))


class TestValidateCloudConfigSchema:
    """Tests for validate_cloudconfig_schema."""

    with_logs = True

    @pytest.mark.parametrize(
        "schema, call_count",
        ((None, 1), ({"properties": {"p1": {"type": "string"}}}, 0)),
    )
    @skipUnlessJsonSchema()
    @mock.patch("cloudinit.config.schema.get_schema")
    def test_validateconfig_schema_use_full_schema_when_no_schema_param(
        self, get_schema, schema, call_count
    ):
        """Use full schema when schema param is absent."""
        get_schema.return_value = {"properties": {"p1": {"type": "string"}}}
        kwargs = {"config": {"p1": "valid"}}
        if schema:
            kwargs["schema"] = schema
        validate_cloudconfig_schema(**kwargs)
        assert call_count == get_schema.call_count

    @skipUnlessJsonSchema()
    def test_validateconfig_schema_non_strict_emits_warnings(self, caplog):
        """When strict is False validate_cloudconfig_schema emits warnings."""
        schema = {"properties": {"p1": {"type": "string"}}}
        validate_cloudconfig_schema({"p1": -1}, schema, strict=False)
        [(module, log_level, log_msg)] = caplog.record_tuples
        assert "cloudinit.config.schema" == module
        assert logging.WARNING == log_level
        assert (
            "Invalid cloud-config provided:\np1: -1 is not of type 'string'"
            == log_msg
        )

    @skipUnlessJsonSchema()
    def test_validateconfig_schema_emits_warning_on_missing_jsonschema(
        self, caplog
    ):
        """Warning from validate_cloudconfig_schema when missing jsonschema."""
        schema = {"properties": {"p1": {"type": "string"}}}
        with mock.patch.dict("sys.modules", **{"jsonschema": ImportError()}):
            validate_cloudconfig_schema({"p1": -1}, schema, strict=True)
        assert "Ignoring schema validation. jsonschema is not present" in (
            caplog.text
        )

    @skipUnlessJsonSchema()
    def test_validateconfig_schema_strict_raises_errors(self):
        """When strict is True validate_cloudconfig_schema raises errors."""
        schema = {"properties": {"p1": {"type": "string"}}}
        with pytest.raises(SchemaValidationError) as context_mgr:
            validate_cloudconfig_schema({"p1": -1}, schema, strict=True)
        assert (
            "Cloud config schema errors: p1: -1 is not of type 'string'"
            == (str(context_mgr.value))
        )

    @skipUnlessJsonSchema()
    def test_validateconfig_schema_honors_formats(self):
        """With strict True, validate_cloudconfig_schema errors on format."""
        schema = {"properties": {"p1": {"type": "string", "format": "email"}}}
        with pytest.raises(SchemaValidationError) as context_mgr:
            validate_cloudconfig_schema({"p1": "-1"}, schema, strict=True)
        assert "Cloud config schema errors: p1: '-1' is not a 'email'" == (
            str(context_mgr.value)
        )

    @skipUnlessJsonSchema()
    def test_validateconfig_schema_honors_formats_strict_metaschema(self):
        """With strict and strict_metaschema True, ensure errors on format"""
        schema = {"properties": {"p1": {"type": "string", "format": "email"}}}
        with pytest.raises(SchemaValidationError) as context_mgr:
            validate_cloudconfig_schema(
                {"p1": "-1"}, schema, strict=True, strict_metaschema=True
            )
        assert "Cloud config schema errors: p1: '-1' is not a 'email'" == str(
            context_mgr.value
        )

    @skipUnlessJsonSchema()
    def test_validateconfig_strict_metaschema_do_not_raise_exception(
        self, caplog
    ):
        """With strict_metaschema=True, do not raise exceptions.

        This flag is currently unused, but is intended for run-time validation.
        This should warn, but not raise.
        """
        schema = {"properties": {"p1": {"types": "string", "format": "email"}}}
        validate_cloudconfig_schema(
            {"p1": "-1"}, schema, strict_metaschema=True
        )
        assert (
            "Meta-schema validation failed, attempting to validate config"
            in caplog.text
        )


class TestCloudConfigExamples:
    metas = get_metas()
    params = [
        (meta["id"], example)
        for meta in metas.values()
        if meta and meta.get("examples")
        for example in meta.get("examples")
    ]

    @pytest.mark.parametrize("schema_id, example", params)
    @skipUnlessJsonSchema()
    def test_validateconfig_schema_of_example(self, schema_id, example):
        """For a given example in a config module we test if it is valid
        according to the unified schema of all config modules
        """
        schema = get_schema()
        config_load = safe_load(example)
        validate_cloudconfig_schema(config_load, schema, strict=True)


class ValidateCloudConfigFileTest(CiTestCase):
    """Tests for validate_cloudconfig_file."""

    def setUp(self):
        super(ValidateCloudConfigFileTest, self).setUp()
        self.config_file = self.tmp_path("cloudcfg.yaml")

    def test_validateconfig_file_error_on_absent_file(self):
        """On absent config_path, validate_cloudconfig_file errors."""
        with self.assertRaises(RuntimeError) as context_mgr:
            validate_cloudconfig_file("/not/here", {})
        self.assertEqual(
            "Configfile /not/here does not exist", str(context_mgr.exception)
        )

    def test_validateconfig_file_error_on_invalid_header(self):
        """On invalid header, validate_cloudconfig_file errors.

        A SchemaValidationError is raised when the file doesn't begin with
        CLOUD_CONFIG_HEADER.
        """
        write_file(self.config_file, "#junk")
        with self.assertRaises(SchemaValidationError) as context_mgr:
            validate_cloudconfig_file(self.config_file, {})
        self.assertEqual(
            "Cloud config schema errors: format-l1.c1: File {0} needs to begin"
            ' with "{1}"'.format(
                self.config_file, CLOUD_CONFIG_HEADER.decode()
            ),
            str(context_mgr.exception),
        )

    def test_validateconfig_file_error_on_non_yaml_scanner_error(self):
        """On non-yaml scan issues, validate_cloudconfig_file errors."""
        # Generate a scanner error by providing text on a single line with
        # improper indent.
        write_file(self.config_file, "#cloud-config\nasdf:\nasdf")
        with self.assertRaises(SchemaValidationError) as context_mgr:
            validate_cloudconfig_file(self.config_file, {})
        self.assertIn(
            "schema errors: format-l3.c1: File {0} is not valid yaml.".format(
                self.config_file
            ),
            str(context_mgr.exception),
        )

    def test_validateconfig_file_error_on_non_yaml_parser_error(self):
        """On non-yaml parser issues, validate_cloudconfig_file errors."""
        write_file(self.config_file, "#cloud-config\n{}}")
        with self.assertRaises(SchemaValidationError) as context_mgr:
            validate_cloudconfig_file(self.config_file, {})
        self.assertIn(
            "schema errors: format-l2.c3: File {0} is not valid yaml.".format(
                self.config_file
            ),
            str(context_mgr.exception),
        )

    @skipUnlessJsonSchema()
    def test_validateconfig_file_sctrictly_validates_schema(self):
        """validate_cloudconfig_file raises errors on invalid schema."""
        schema = {"properties": {"p1": {"type": "string", "format": "string"}}}
        write_file(self.config_file, "#cloud-config\np1: -1")
        with self.assertRaises(SchemaValidationError) as context_mgr:
            validate_cloudconfig_file(self.config_file, schema)
        self.assertEqual(
            "Cloud config schema errors: p1: -1 is not of type 'string'",
            str(context_mgr.exception),
        )


class GetSchemaDocTest(CiTestCase):
    """Tests for get_meta_doc."""

    def setUp(self):
        super(GetSchemaDocTest, self).setUp()
        self.required_schema = {
            "title": "title",
            "description": "description",
            "id": "id",
            "name": "name",
            "frequency": "frequency",
            "distros": ["debian", "rhel"],
        }
        self.meta: MetaSchema = {
            "title": "title",
            "description": "description",
            "id": "id",
            "name": "name",
            "frequency": "frequency",
            "distros": ["debian", "rhel"],
            "examples": [
                'ex1:\n    [don\'t, expand, "this"]',
                "ex2: true",
            ],
        }

    def test_get_meta_doc_returns_restructured_text(self):
        """get_meta_doc returns restructured text for a cloudinit schema."""
        full_schema = copy(self.required_schema)
        full_schema.update(
            {
                "properties": {
                    "prop1": {
                        "type": "array",
                        "description": "prop-description",
                        "items": {"type": "integer"},
                    }
                }
            }
        )

        doc = get_meta_doc(self.meta, full_schema)
        self.assertEqual(
            dedent(
                """
                name
                ----
                **Summary:** title

                description

                **Internal name:** ``id``

                **Module frequency:** frequency

                **Supported distros:** debian, rhel

                **Config schema**:
                    **prop1:** (array of integer) prop-description

                **Examples**::

                    ex1:
                        [don't, expand, "this"]
                    # --- Example2 ---
                    ex2: true
            """
            ),
            doc,
        )

    def test_get_meta_doc_handles_multiple_types(self):
        """get_meta_doc delimits multiple property types with a '/'."""
        schema = {"properties": {"prop1": {"type": ["string", "integer"]}}}
        self.assertIn(
            "**prop1:** (string/integer)", get_meta_doc(self.meta, schema)
        )

    def test_get_meta_doc_handles_enum_types(self):
        """get_meta_doc converts enum types to yaml and delimits with '/'."""
        schema = {"properties": {"prop1": {"enum": [True, False, "stuff"]}}}
        self.assertIn(
            "**prop1:** (true/false/stuff)", get_meta_doc(self.meta, schema)
        )

    def test_get_meta_doc_handles_nested_oneof_property_types(self):
        """get_meta_doc describes array items oneOf declarations in type."""
        schema = {
            "properties": {
                "prop1": {
                    "type": "array",
                    "items": {
                        "oneOf": [{"type": "string"}, {"type": "integer"}]
                    },
                }
            }
        }
        self.assertIn(
            "**prop1:** (array of (string)/(integer))",
            get_meta_doc(self.meta, schema),
        )

    def test_get_meta_doc_handles_string_examples(self):
        """get_meta_doc properly indented examples as a list of strings."""
        full_schema = copy(self.required_schema)
        full_schema.update(
            {
                "examples": [
                    'ex1:\n    [don\'t, expand, "this"]',
                    "ex2: true",
                ],
                "properties": {
                    "prop1": {
                        "type": "array",
                        "description": "prop-description",
                        "items": {"type": "integer"},
                    }
                },
            }
        )
        self.assertIn(
            dedent(
                """
                **Config schema**:
                    **prop1:** (array of integer) prop-description

                **Examples**::

                    ex1:
                        [don't, expand, "this"]
                    # --- Example2 ---
                    ex2: true
            """
            ),
            get_meta_doc(self.meta, full_schema),
        )

    def test_get_meta_doc_properly_parse_description(self):
        """get_meta_doc description properly formatted"""
        schema = {
            "properties": {
                "p1": {
                    "type": "string",
                    "description": dedent(
                        """\
                        This item
                        has the
                        following options:

                          - option1
                          - option2
                          - option3

                        The default value is
                        option1"""
                    ),
                }
            }
        }

        self.assertIn(
            dedent(
                """
                **Config schema**:
                    **p1:** (string) This item has the following options:

                            - option1
                            - option2
                            - option3

                    The default value is option1

            """
            ),
            get_meta_doc(self.meta, schema),
        )

    def test_get_meta_doc_raises_key_errors(self):
        """get_meta_doc raises KeyErrors on missing keys."""
        schema = {
            "properties": {
                "prop1": {
                    "type": "array",
                    "items": {
                        "oneOf": [{"type": "string"}, {"type": "integer"}]
                    },
                }
            }
        }
        for key in self.meta:
            invalid_meta = copy(self.meta)
            invalid_meta.pop(key)
            with self.assertRaises(KeyError) as context_mgr:
                get_meta_doc(invalid_meta, schema)
            self.assertIn(key, str(context_mgr.exception))

    def test_label_overrides_property_name(self):
        """get_meta_doc overrides property name with label."""
        schema = {
            "properties": {
                "prop1": {
                    "type": "string",
                    "label": "label1",
                },
                "prop_no_label": {
                    "type": "string",
                },
                "prop_array": {
                    "label": "array_label",
                    "type": "array",
                    "items": {
                        "type": "object",
                        "properties": {
                            "some_prop": {"type": "number"},
                        },
                    },
                },
            },
            "patternProperties": {
                "^.*$": {
                    "type": "string",
                    "label": "label2",
                }
            },
        }
        meta_doc = get_meta_doc(self.meta, schema)
        assert "**label1:** (string)" in meta_doc
        assert "**label2:** (string" in meta_doc
        assert "**prop_no_label:** (string)" in meta_doc
        assert "Each item in **array_label** list" in meta_doc

        assert "prop1" not in meta_doc
        assert ".*" not in meta_doc


class AnnotatedCloudconfigFileTest(CiTestCase):
    maxDiff = None

    def test_annotated_cloudconfig_file_no_schema_errors(self):
        """With no schema_errors, print the original content."""
        content = b"ntp:\n  pools: [ntp1.pools.com]\n"
        self.assertEqual(
            content, annotated_cloudconfig_file({}, content, schema_errors=[])
        )

    def test_annotated_cloudconfig_file_with_non_dict_cloud_config(self):
        """Error when empty non-dict cloud-config is provided.

        OurJSON validation when user-data is None type generates a bunch
        schema validation errors of the format:
        ('', "None is not of type 'object'"). Ignore those symptoms and
        report the general problem instead.
        """
        content = b"\n\n\n"
        expected = "\n".join(
            [
                content.decode(),
                "# Errors: -------------",
                "# E1: Cloud-config is not a YAML dict.\n\n",
            ]
        )
        self.assertEqual(
            expected,
            annotated_cloudconfig_file(
                None,
                content,
                schema_errors=[("", "None is not of type 'object'")],
            ),
        )

    def test_annotated_cloudconfig_file_schema_annotates_and_adds_footer(self):
        """With schema_errors, error lines are annotated and a footer added."""
        content = dedent(
            """\
            #cloud-config
            # comment
            ntp:
              pools: [-99, 75]
            """
        ).encode()
        expected = dedent(
            """\
            #cloud-config
            # comment
            ntp:		# E1
              pools: [-99, 75]		# E2,E3

            # Errors: -------------
            # E1: Some type error
            # E2: -99 is not a string
            # E3: 75 is not a string

            """
        )
        parsed_config = safe_load(content[13:])
        schema_errors = [
            ("ntp", "Some type error"),
            ("ntp.pools.0", "-99 is not a string"),
            ("ntp.pools.1", "75 is not a string"),
        ]
        self.assertEqual(
            expected,
            annotated_cloudconfig_file(parsed_config, content, schema_errors),
        )

    def test_annotated_cloudconfig_file_annotates_separate_line_items(self):
        """Errors are annotated for lists with items on separate lines."""
        content = dedent(
            """\
            #cloud-config
            # comment
            ntp:
              pools:
                - -99
                - 75
            """
        ).encode()
        expected = dedent(
            """\
            ntp:
              pools:
                - -99		# E1
                - 75		# E2
            """
        )
        parsed_config = safe_load(content[13:])
        schema_errors = [
            ("ntp.pools.0", "-99 is not a string"),
            ("ntp.pools.1", "75 is not a string"),
        ]
        self.assertIn(
            expected,
            annotated_cloudconfig_file(parsed_config, content, schema_errors),
        )


class TestMain:

    exclusive_combinations = itertools.combinations(
        ["--system", "--docs all", "--config-file something"], 2
    )

    @pytest.mark.parametrize("params", exclusive_combinations)
    def test_main_exclusive_args(self, params, capsys):
        """Main exits non-zero and error on required exclusive args."""
        params = list(itertools.chain(*[a.split() for a in params]))
        with mock.patch("sys.argv", ["mycmd"] + params):
            with pytest.raises(SystemExit) as context_manager:
                main()
        assert 1 == context_manager.value.code

        _out, err = capsys.readouterr()
        expected = (
            "Error:\n"
            "Expected one of --config-file, --system or --docs arguments\n"
        )
        assert expected == err

    def test_main_missing_args(self, capsys):
        """Main exits non-zero and reports an error on missing parameters."""
        with mock.patch("sys.argv", ["mycmd"]):
            with pytest.raises(SystemExit) as context_manager:
                main()
        assert 1 == context_manager.value.code

        _out, err = capsys.readouterr()
        expected = (
            "Error:\n"
            "Expected one of --config-file, --system or --docs arguments\n"
        )
        assert expected == err

    def test_main_absent_config_file(self, capsys):
        """Main exits non-zero when config file is absent."""
        myargs = ["mycmd", "--annotate", "--config-file", "NOT_A_FILE"]
        with mock.patch("sys.argv", myargs):
            with pytest.raises(SystemExit) as context_manager:
                main()
        assert 1 == context_manager.value.code
        _out, err = capsys.readouterr()
        assert "Error:\nConfigfile NOT_A_FILE does not exist\n" == err

    def test_main_invalid_flag_combo(self, capsys):
        """Main exits non-zero when invalid flag combo used."""
        myargs = ["mycmd", "--annotate", "--docs", "DOES_NOT_MATTER"]
        with mock.patch("sys.argv", myargs):
            with pytest.raises(SystemExit) as context_manager:
                main()
        assert 1 == context_manager.value.code
        _, err = capsys.readouterr()
        assert (
            "Error:\nInvalid flag combination. "
            "Cannot use --annotate with --docs\n" == err
        )

    def test_main_prints_docs(self, capsys):
        """When --docs parameter is provided, main generates documentation."""
        myargs = ["mycmd", "--docs", "all"]
        with mock.patch("sys.argv", myargs):
            assert 0 == main(), "Expected 0 exit code"
        out, _err = capsys.readouterr()
        assert "\nNTP\n---\n" in out
        assert "\nRuncmd\n------\n" in out

    def test_main_validates_config_file(self, tmpdir, capsys):
        """When --config-file parameter is provided, main validates schema."""
        myyaml = tmpdir.join("my.yaml")
        myargs = ["mycmd", "--config-file", myyaml.strpath]
        myyaml.write(b"#cloud-config\nntp:")  # shortest ntp schema
        with mock.patch("sys.argv", myargs):
            assert 0 == main(), "Expected 0 exit code"
        out, _err = capsys.readouterr()
        assert "Valid cloud-config: {0}\n".format(myyaml) == out

    @mock.patch("cloudinit.config.schema.read_cfg_paths")
    @mock.patch("cloudinit.config.schema.os.getuid", return_value=0)
    def test_main_validates_system_userdata(
        self, m_getuid, m_read_cfg_paths, capsys, paths
    ):
        """When --system is provided, main validates system userdata."""
        m_read_cfg_paths.return_value = paths
        ud_file = paths.get_ipath_cur("userdata_raw")
        write_file(ud_file, b"#cloud-config\nntp:")
        myargs = ["mycmd", "--system"]
        with mock.patch("sys.argv", myargs):
            assert 0 == main(), "Expected 0 exit code"
        out, _err = capsys.readouterr()
        assert "Valid cloud-config: system userdata\n" == out

    @mock.patch("cloudinit.config.schema.os.getuid", return_value=1000)
    def test_main_system_userdata_requires_root(self, m_getuid, capsys, paths):
        """Non-root user can't use --system param"""
        myargs = ["mycmd", "--system"]
        with mock.patch("sys.argv", myargs):
            with pytest.raises(SystemExit) as context_manager:
                main()
        assert 1 == context_manager.value.code
        _out, err = capsys.readouterr()
        expected = (
            "Error:\nUnable to read system userdata as non-root user. "
            "Try using sudo\n"
        )
        assert expected == err


def _get_meta_doc_examples():
    examples_dir = Path(cloud_init_project_dir("doc/examples"))
    assert examples_dir.is_dir()

    return (
        str(f)
        for f in examples_dir.glob("cloud-config*.txt")
        if not f.name.startswith("cloud-config-archive")
    )


class TestSchemaDocExamples:
    schema = get_schema()

    @pytest.mark.parametrize("example_path", _get_meta_doc_examples())
    @skipUnlessJsonSchema()
    def test_schema_doc_examples(self, example_path):
        validate_cloudconfig_file(example_path, self.schema)


class TestStrictMetaschema:
    """Validate that schemas follow a stricter metaschema definition than
    the default. This disallows arbitrary key/value pairs.
    """

    @skipUnlessJsonSchema()
    def test_modules(self):
        """Validate all modules with a stricter metaschema"""
        (validator, _) = get_jsonschema_validator()
        for (name, value) in get_schemas().items():
            if value:
                validate_cloudconfig_metaschema(validator, value)
            else:
                logging.warning("module %s has no schema definition", name)

    @skipUnlessJsonSchema()
    def test_validate_bad_module(self):
        """Throw exception by default, don't throw if throw=False

        item should be 'items' and is therefore interpreted as an additional
        property which is invalid with a strict metaschema
        """
        (validator, _) = get_jsonschema_validator()
        schema = {
            "type": "array",
            "item": {
                "type": "object",
            },
        }
        with pytest.raises(
            SchemaValidationError,
            match=r"Additional properties are not allowed.*",
        ):

            validate_cloudconfig_metaschema(validator, schema)

        validate_cloudconfig_metaschema(validator, schema, throw=False)


# vi: ts=4 expandtab syntax=python



# Author: Antti Myyrä <antti.myyra@upcloud.com>
#
# This file is part of cloud-init. See LICENSE file for license information.

# UpCloud server metadata API:
# https://developers.upcloud.com/1.3/8-servers/#metadata-service

from cloudinit import log as logging
from cloudinit import net as cloudnet
from cloudinit import sources, util
from cloudinit.net.dhcp import EphemeralDHCPv4, NoDHCPLeaseError
from cloudinit.sources.helpers import upcloud as uc_helper

LOG = logging.getLogger(__name__)

BUILTIN_DS_CONFIG = {"metadata_url": "http://169.254.169.254/metadata/v1.json"}

# Wait for a up to a minute, retrying the meta-data server
# every 2 seconds.
MD_RETRIES = 30
MD_TIMEOUT = 2
MD_WAIT_RETRY = 2


class DataSourceUpCloud(sources.DataSource):

    dsname = "UpCloud"

    # We'll perform DHCP setup only in init-local, see DataSourceUpCloudLocal
    perform_dhcp_setup = False

    def __init__(self, sys_cfg, distro, paths):
        sources.DataSource.__init__(self, sys_cfg, distro, paths)
        self.distro = distro
        self.metadata = dict()
        self.ds_cfg = util.mergemanydict(
            [
                util.get_cfg_by_path(sys_cfg, ["datasource", "UpCloud"], {}),
                BUILTIN_DS_CONFIG,
            ]
        )
        self.metadata_address = self.ds_cfg["metadata_url"]
        self.retries = self.ds_cfg.get("retries", MD_RETRIES)
        self.timeout = self.ds_cfg.get("timeout", MD_TIMEOUT)
        self.wait_retry = self.ds_cfg.get("wait_retry", MD_WAIT_RETRY)
        self._network_config = None

    def _get_sysinfo(self):
        return uc_helper.read_sysinfo()

    def _read_metadata(self):
        return uc_helper.read_metadata(
            self.metadata_address,
            timeout=self.timeout,
            sec_between=self.wait_retry,
            retries=self.retries,
        )

    def _get_data(self):
        (is_upcloud, server_uuid) = self._get_sysinfo()

        # only proceed if we know we are on UpCloud
        if not is_upcloud:
            return False

        LOG.info("Running on UpCloud. server_uuid=%s", server_uuid)

        if self.perform_dhcp_setup:  # Setup networking in init-local stage.
            try:
                LOG.debug("Finding a fallback NIC")
                nic = cloudnet.find_fallback_nic()
                LOG.debug("Discovering metadata via DHCP interface %s", nic)
                with EphemeralDHCPv4(nic):
                    md = util.log_time(
                        logfunc=LOG.debug,
                        msg="Reading from metadata service",
                        func=self._read_metadata,
                    )
            except (NoDHCPLeaseError, sources.InvalidMetaDataException) as e:
                util.logexc(LOG, str(e))
                return False
        else:
            try:
                LOG.debug(
                    "Discovering metadata without DHCP-configured networking"
                )
                md = util.log_time(
                    logfunc=LOG.debug,
                    msg="Reading from metadata service",
                    func=self._read_metadata,
                )
            except sources.InvalidMetaDataException as e:
                util.logexc(LOG, str(e))
                LOG.info(
                    "No DHCP-enabled interfaces available, "
                    "unable to fetch metadata for %s",
                    server_uuid,
                )
                return False

        self.metadata_full = md
        self.metadata["instance-id"] = md.get("instance_id", server_uuid)
        self.metadata["local-hostname"] = md.get("hostname")
        self.metadata["network"] = md.get("network")
        self.metadata["public-keys"] = md.get("public_keys")
        self.metadata["availability_zone"] = md.get("region", "default")
        self.vendordata_raw = md.get("vendor_data", None)
        self.userdata_raw = md.get("user_data", None)

        return True

    def check_instance_id(self, sys_cfg):
        return sources.instance_id_matches_system_uuid(self.get_instance_id())

    @property
    def network_config(self):
        """
        Configure the networking. This needs to be done each boot,
        since the IP and interface information might have changed
        due to reconfiguration.
        """

        if self._network_config:
            return self._network_config

        raw_network_config = self.metadata.get("network")
        if not raw_network_config:
            raise Exception("Unable to get network meta-data from server....")

        self._network_config = uc_helper.convert_network_config(
            raw_network_config,
        )

        return self._network_config


class DataSourceUpCloudLocal(DataSourceUpCloud):
    """
    Run in init-local using a DHCP discovery prior to metadata crawl.

    In init-local, no network is available. This subclass sets up minimal
    networking with dhclient on a viable nic so that it can talk to the
    metadata service. If the metadata service provides network configuration
    then render the network configuration for that instance based on metadata.
    """

    perform_dhcp_setup = True  # Get metadata network config if present


# Used to match classes to dependencies
datasources = [
    (DataSourceUpCloudLocal, (sources.DEP_FILESYSTEM,)),
    (DataSourceUpCloud, (sources.DEP_FILESYSTEM, sources.DEP_NETWORK)),
]


# Return a list of data sources that match this set of dependencies
def get_datasource_list(depends):
    return sources.list_from_depends(depends, datasources)


# vi: ts=4 expandtab



# Copyright (C) 2017 Canonical Ltd.
#
# This file is part of cloud-init. See LICENSE file for license information.

"""Cloud-init apport interface"""

try:
    from apport.hookutils import (
        attach_file,
        attach_root_command_outputs,
        root_command_output,
    )

    has_apport = True
except ImportError:
    has_apport = False


KNOWN_CLOUD_NAMES = [
    "AliYun",
    "AltCloud",
    "Amazon - Ec2",
    "Azure",
    "Bigstep",
    "Brightbox",
    "CloudSigma",
    "CloudStack",
    "DigitalOcean",
    "E24Cloud",
    "GCE - Google Compute Engine",
    "Exoscale",
    "Hetzner Cloud",
    "IBM - (aka SoftLayer or BlueMix)",
    "LXD",
    "MAAS",
    "NoCloud",
    "OpenNebula",
    "OpenStack",
    "Oracle",
    "OVF",
    "RbxCloud - (HyperOne, Rootbox, Rubikon)",
    "OpenTelekomCloud",
    "SAP Converged Cloud",
    "Scaleway",
    "SmartOS",
    "UpCloud",
    "VMware",
    "Vultr",
    "ZStack",
    "Other",
]

# Potentially clear text collected logs
CLOUDINIT_LOG = "/var/log/cloud-init.log"
CLOUDINIT_OUTPUT_LOG = "/var/log/cloud-init-output.log"
USER_DATA_FILE = "/var/lib/cloud/instance/user-data.txt"  # Optional


def attach_cloud_init_logs(report, ui=None):
    """Attach cloud-init logs and tarfile from 'cloud-init collect-logs'."""
    attach_root_command_outputs(
        report,
        {
            "cloud-init-log-warnings": (
                'egrep -i "warn|error" /var/log/cloud-init.log'
            ),
            "cloud-init-output.log.txt": "cat /var/log/cloud-init-output.log",
        },
    )
    root_command_output(
        ["cloud-init", "collect-logs", "-t", "/tmp/cloud-init-logs.tgz"]
    )
    attach_file(report, "/tmp/cloud-init-logs.tgz", "logs.tgz")


def attach_hwinfo(report, ui=None):
    """Optionally attach hardware info from lshw."""
    prompt = (
        "Your device details (lshw) may be useful to developers when"
        " addressing this bug, but gathering it requires admin privileges."
        " Would you like to include this info?"
    )
    if ui and ui.yesno(prompt):
        attach_root_command_outputs(report, {"lshw.txt": "lshw"})


def attach_cloud_info(report, ui=None):
    """Prompt for cloud details if available."""
    if ui:
        prompt = "Is this machine running in a cloud environment?"
        response = ui.yesno(prompt)
        if response is None:
            raise StopIteration  # User cancelled
        if response:
            prompt = (
                "Please select the cloud vendor or environment in which"
                " this instance is running"
            )
            response = ui.choice(prompt, KNOWN_CLOUD_NAMES)
            if response:
                report["CloudName"] = KNOWN_CLOUD_NAMES[response[0]]
            else:
                report["CloudName"] = "None"


def attach_user_data(report, ui=None):
    """Optionally provide user-data if desired."""
    if ui:
        prompt = (
            "Your user-data or cloud-config file can optionally be provided"
            " from {0} and could be useful to developers when addressing this"
            " bug. Do you wish to attach user-data to this bug?".format(
                USER_DATA_FILE
            )
        )
        response = ui.yesno(prompt)
        if response is None:
            raise StopIteration  # User cancelled
        if response:
            attach_file(report, USER_DATA_FILE, "user_data.txt")


def add_bug_tags(report):
    """Add any appropriate tags to the bug."""
    if "JournalErrors" in report.keys():
        errors = report["JournalErrors"]
        if "Breaking ordering cycle" in errors:
            report["Tags"] = "systemd-ordering"


def add_info(report, ui):
    """This is an entry point to run cloud-init's apport functionality.

    Distros which want apport support will have a cloud-init package-hook at
    /usr/share/apport/package-hooks/cloud-init.py which defines an add_info
    function and returns the result of cloudinit.apport.add_info(report, ui).
    """
    if not has_apport:
        raise RuntimeError(
            "No apport imports discovered. Apport functionality disabled"
        )
    attach_cloud_init_logs(report, ui)
    attach_hwinfo(report, ui)
    attach_cloud_info(report, ui)
    attach_user_data(report, ui)
    add_bug_tags(report)
    return True


# vi: ts=4 expandtab



# This file is part of cloud-init. See LICENSE file for license information.

import logging
import os
import shutil
import tempfile

from cloudinit import handlers, helpers, settings, url_helper, util
from cloudinit.cmd import main
from tests.unittests.helpers import CiTestCase, ExitStack, TestCase, mock


class FakeModule(handlers.Handler):
    def __init__(self):
        handlers.Handler.__init__(self, settings.PER_ALWAYS)
        self.types = []

    def list_types(self):
        return self.types

    def handle_part(self, data, ctype, filename, payload, frequency):
        pass


class TestWalkerHandleHandler(TestCase):
    def setUp(self):
        super(TestWalkerHandleHandler, self).setUp()
        tmpdir = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, tmpdir)

        self.data = {
            "handlercount": 0,
            "frequency": "",
            "handlerdir": tmpdir,
            "handlers": helpers.ContentHandlers(),
            "data": None,
        }

        self.expected_module_name = "part-handler-%03d" % (
            self.data["handlercount"],
        )
        expected_file_name = "%s.py" % self.expected_module_name
        self.expected_file_fullname = os.path.join(
            self.data["handlerdir"], expected_file_name
        )
        self.module_fake = FakeModule()
        self.ctype = None
        self.filename = None
        self.payload = "dummy payload"

        # Mock the write_file() function.  We'll assert that it got called as
        # expected in each of the individual tests.
        resources = ExitStack()
        self.addCleanup(resources.close)
        self.write_file_mock = resources.enter_context(
            mock.patch("cloudinit.util.write_file")
        )

    def test_no_errors(self):
        """Payload gets written to file and added to C{pdata}."""
        with mock.patch(
            "cloudinit.importer.import_module", return_value=self.module_fake
        ) as mockobj:
            handlers.walker_handle_handler(
                self.data, self.ctype, self.filename, self.payload
            )
            mockobj.assert_called_once_with(self.expected_module_name)
        self.write_file_mock.assert_called_once_with(
            self.expected_file_fullname, self.payload, 0o600
        )
        self.assertEqual(self.data["handlercount"], 1)

    def test_import_error(self):
        """Module import errors are logged. No handler added to C{pdata}."""
        with mock.patch(
            "cloudinit.importer.import_module", side_effect=ImportError
        ) as mockobj:
            handlers.walker_handle_handler(
                self.data, self.ctype, self.filename, self.payload
            )
            mockobj.assert_called_once_with(self.expected_module_name)
        self.write_file_mock.assert_called_once_with(
            self.expected_file_fullname, self.payload, 0o600
        )
        self.assertEqual(self.data["handlercount"], 0)

    def test_attribute_error(self):
        """Attribute errors are logged. No handler added to C{pdata}."""
        with mock.patch(
            "cloudinit.importer.import_module",
            side_effect=AttributeError,
            return_value=self.module_fake,
        ) as mockobj:
            handlers.walker_handle_handler(
                self.data, self.ctype, self.filename, self.payload
            )
            mockobj.assert_called_once_with(self.expected_module_name)
        self.write_file_mock.assert_called_once_with(
            self.expected_file_fullname, self.payload, 0o600
        )
        self.assertEqual(self.data["handlercount"], 0)


class TestHandlerHandlePart(TestCase):
    def setUp(self):
        super(TestHandlerHandlePart, self).setUp()
        self.data = "fake data"
        self.ctype = "fake ctype"
        self.filename = "fake filename"
        self.payload = "fake payload"
        self.frequency = settings.PER_INSTANCE
        self.headers = {
            "Content-Type": self.ctype,
        }

    def test_normal_version_1(self):
        """
        C{handle_part} is called without C{frequency} for
        C{handler_version} == 1.
        """
        mod_mock = mock.Mock(
            frequency=settings.PER_INSTANCE, handler_version=1
        )
        handlers.run_part(
            mod_mock,
            self.data,
            self.filename,
            self.payload,
            self.frequency,
            self.headers,
        )
        # Assert that the handle_part() method of the mock object got
        # called with the expected arguments.
        mod_mock.handle_part.assert_called_once_with(
            self.data, self.ctype, self.filename, self.payload
        )

    def test_normal_version_2(self):
        """
        C{handle_part} is called with C{frequency} for
        C{handler_version} == 2.
        """
        mod_mock = mock.Mock(
            frequency=settings.PER_INSTANCE, handler_version=2
        )
        handlers.run_part(
            mod_mock,
            self.data,
            self.filename,
            self.payload,
            self.frequency,
            self.headers,
        )
        # Assert that the handle_part() method of the mock object got
        # called with the expected arguments.
        mod_mock.handle_part.assert_called_once_with(
            self.data,
            self.ctype,
            self.filename,
            self.payload,
            settings.PER_INSTANCE,
        )

    def test_modfreq_per_always(self):
        """
        C{handle_part} is called regardless of frequency if nofreq is always.
        """
        self.frequency = "once"
        mod_mock = mock.Mock(frequency=settings.PER_ALWAYS, handler_version=1)
        handlers.run_part(
            mod_mock,
            self.data,
            self.filename,
            self.payload,
            self.frequency,
            self.headers,
        )
        # Assert that the handle_part() method of the mock object got
        # called with the expected arguments.
        mod_mock.handle_part.assert_called_once_with(
            self.data, self.ctype, self.filename, self.payload
        )

    def test_no_handle_when_modfreq_once(self):
        """C{handle_part} is not called if frequency is once."""
        self.frequency = "once"
        mod_mock = mock.Mock(frequency=settings.PER_ONCE)
        handlers.run_part(
            mod_mock,
            self.data,
            self.filename,
            self.payload,
            self.frequency,
            self.headers,
        )
        self.assertEqual(0, mod_mock.handle_part.call_count)

    def test_exception_is_caught(self):
        """Exceptions within C{handle_part} are caught and logged."""
        mod_mock = mock.Mock(
            frequency=settings.PER_INSTANCE, handler_version=1
        )
        mod_mock.handle_part.side_effect = Exception
        try:
            handlers.run_part(
                mod_mock,
                self.data,
                self.filename,
                self.payload,
                self.frequency,
                self.headers,
            )
        except Exception:
            self.fail("Exception was not caught in handle_part")

        mod_mock.handle_part.assert_called_once_with(
            self.data, self.ctype, self.filename, self.payload
        )


class TestCmdlineUrl(CiTestCase):
    def test_parse_cmdline_url_nokey_raises_keyerror(self):
        self.assertRaises(
            KeyError, main.parse_cmdline_url, "root=foo bar single"
        )

    def test_parse_cmdline_url_found(self):
        cmdline = "root=foo bar single url=http://example.com arg1 -v"
        self.assertEqual(
            ("url", "http://example.com"), main.parse_cmdline_url(cmdline)
        )

    @mock.patch("cloudinit.cmd.main.url_helper.read_file_or_url")
    def test_invalid_content(self, m_read):
        key = "cloud-config-url"
        url = "http://example.com/foo"
        cmdline = "ro %s=%s bar=1" % (key, url)
        m_read.return_value = url_helper.StringResponse(b"unexpected blob")

        fpath = self.tmp_path("ccfile")
        lvl, msg = main.attempt_cmdline_url(
            fpath, network=True, cmdline=cmdline
        )
        self.assertEqual(logging.WARN, lvl)
        self.assertIn(url, msg)
        self.assertFalse(os.path.exists(fpath))

    @mock.patch("cloudinit.cmd.main.url_helper.read_file_or_url")
    def test_valid_content(self, m_read):
        url = "http://example.com/foo"
        payload = b"#cloud-config\nmydata: foo\nbar: wark\n"
        cmdline = "ro %s=%s bar=1" % ("cloud-config-url", url)

        m_read.return_value = url_helper.StringResponse(payload)
        fpath = self.tmp_path("ccfile")
        lvl, msg = main.attempt_cmdline_url(
            fpath, network=True, cmdline=cmdline
        )
        self.assertEqual(util.load_file(fpath, decode=False), payload)
        self.assertEqual(logging.INFO, lvl)
        self.assertIn(url, msg)

    @mock.patch("cloudinit.cmd.main.url_helper.read_file_or_url")
    def test_no_key_found(self, m_read):
        cmdline = "ro mykey=http://example.com/foo root=foo"
        fpath = self.tmp_path("ccpath")
        lvl, _msg = main.attempt_cmdline_url(
            fpath, network=True, cmdline=cmdline
        )

        m_read.assert_not_called()
        self.assertFalse(os.path.exists(fpath))
        self.assertEqual(logging.DEBUG, lvl)

    @mock.patch("cloudinit.cmd.main.url_helper.read_file_or_url")
    def test_exception_warns(self, m_read):
        url = "http://example.com/foo"
        cmdline = "ro cloud-config-url=%s root=LABEL=bar" % url
        fpath = self.tmp_path("ccfile")
        m_read.side_effect = url_helper.UrlError(
            cause="Unexpected Error", url="http://example.com/foo"
        )

        lvl, msg = main.attempt_cmdline_url(
            fpath, network=True, cmdline=cmdline
        )
        self.assertEqual(logging.WARN, lvl)
        self.assertIn(url, msg)
        self.assertFalse(os.path.exists(fpath))


# vi: ts=4 expandtab



"""Integration test for output redirection.

This test redirects the output of a command to a file and then checks the file.

(This is ported from
``tests/cloud_tests/testcases/main/command_output_simple.yaml``.)"""
import pytest

from tests.integration_tests.instances import IntegrationInstance

USER_DATA = """\
#cloud-config
output: { all: "| tee -a /var/log/cloud-init-test-output" }
final_message: "should be last line in cloud-init-test-output file"
"""


@pytest.mark.user_data(USER_DATA)
def test_runcmd(client: IntegrationInstance):
    log = client.read_from_file("/var/log/cloud-init-test-output")
    assert "should be last line in cloud-init-test-output file" in log



# This file is part of cloud-init. See LICENSE file for license information.

from unittest import mock

from cloudinit import distros, helpers, settings
from cloudinit.distros import ug_util
from tests.unittests.helpers import TestCase

bcfg = {
    "name": "bob",
    "plain_text_passwd": "ubuntu",
    "home": "/home/ubuntu",
    "shell": "/bin/bash",
    "lock_passwd": True,
    "gecos": "Ubuntu",
    "groups": ["foo"],
}


class TestUGNormalize(TestCase):
    def setUp(self):
        super(TestUGNormalize, self).setUp()
        self.add_patch("cloudinit.util.system_is_snappy", "m_snappy")

    def _make_distro(self, dtype, def_user=None):
        cfg = dict(settings.CFG_BUILTIN)
        cfg["system_info"]["distro"] = dtype
        paths = helpers.Paths(cfg["system_info"]["paths"])
        distro_cls = distros.fetch(dtype)
        if def_user:
            cfg["system_info"]["default_user"] = def_user.copy()
        distro = distro_cls(dtype, cfg["system_info"], paths)
        return distro

    def _norm(self, cfg, distro):
        return ug_util.normalize_users_groups(cfg, distro)

    def test_group_dict(self):
        distro = self._make_distro("ubuntu")
        g = {
            "groups": [
                {"ubuntu": ["foo", "bar"], "bob": "users"},
                "cloud-users",
                {"bob": "users2"},
            ]
        }
        (_users, groups) = self._norm(g, distro)
        self.assertIn("ubuntu", groups)
        ub_members = groups["ubuntu"]
        self.assertEqual(sorted(["foo", "bar"]), sorted(ub_members))
        self.assertIn("bob", groups)
        b_members = groups["bob"]
        self.assertEqual(sorted(["users", "users2"]), sorted(b_members))

    def test_basic_groups(self):
        distro = self._make_distro("ubuntu")
        ug_cfg = {
            "groups": ["bob"],
        }
        (users, groups) = self._norm(ug_cfg, distro)
        self.assertIn("bob", groups)
        self.assertEqual({}, users)

    def test_csv_groups(self):
        distro = self._make_distro("ubuntu")
        ug_cfg = {
            "groups": "bob,joe,steve",
        }
        (users, groups) = self._norm(ug_cfg, distro)
        self.assertIn("bob", groups)
        self.assertIn("joe", groups)
        self.assertIn("steve", groups)
        self.assertEqual({}, users)

    def test_more_groups(self):
        distro = self._make_distro("ubuntu")
        ug_cfg = {"groups": ["bob", "joe", "steve"]}
        (users, groups) = self._norm(ug_cfg, distro)
        self.assertIn("bob", groups)
        self.assertIn("joe", groups)
        self.assertIn("steve", groups)
        self.assertEqual({}, users)

    def test_member_groups(self):
        distro = self._make_distro("ubuntu")
        ug_cfg = {
            "groups": {
                "bob": ["s"],
                "joe": [],
                "steve": [],
            }
        }
        (users, groups) = self._norm(ug_cfg, distro)
        self.assertIn("bob", groups)
        self.assertEqual(["s"], groups["bob"])
        self.assertEqual([], groups["joe"])
        self.assertIn("joe", groups)
        self.assertIn("steve", groups)
        self.assertEqual({}, users)

    def test_users_simple_dict(self):
        distro = self._make_distro("ubuntu", bcfg)
        ug_cfg = {
            "users": {
                "default": True,
            }
        }
        (users, _groups) = self._norm(ug_cfg, distro)
        self.assertIn("bob", users)
        ug_cfg = {
            "users": {
                "default": "yes",
            }
        }
        (users, _groups) = self._norm(ug_cfg, distro)
        self.assertIn("bob", users)
        ug_cfg = {
            "users": {
                "default": "1",
            }
        }
        (users, _groups) = self._norm(ug_cfg, distro)
        self.assertIn("bob", users)

    def test_users_simple_dict_no(self):
        distro = self._make_distro("ubuntu", bcfg)
        ug_cfg = {
            "users": {
                "default": False,
            }
        }
        (users, _groups) = self._norm(ug_cfg, distro)
        self.assertEqual({}, users)
        ug_cfg = {
            "users": {
                "default": "no",
            }
        }
        (users, _groups) = self._norm(ug_cfg, distro)
        self.assertEqual({}, users)

    def test_users_simple_csv(self):
        distro = self._make_distro("ubuntu")
        ug_cfg = {
            "users": "joe,bob",
        }
        (users, _groups) = self._norm(ug_cfg, distro)
        self.assertIn("joe", users)
        self.assertIn("bob", users)
        self.assertEqual({"default": False}, users["joe"])
        self.assertEqual({"default": False}, users["bob"])

    def test_users_simple(self):
        distro = self._make_distro("ubuntu")
        ug_cfg = {
            "users": ["joe", "bob"],
        }
        (users, _groups) = self._norm(ug_cfg, distro)
        self.assertIn("joe", users)
        self.assertIn("bob", users)
        self.assertEqual({"default": False}, users["joe"])
        self.assertEqual({"default": False}, users["bob"])

    def test_users_old_user(self):
        distro = self._make_distro("ubuntu", bcfg)
        ug_cfg = {"user": "zetta", "users": "default"}
        (users, _groups) = self._norm(ug_cfg, distro)
        self.assertNotIn("bob", users)  # Bob is not the default now, zetta is
        self.assertIn("zetta", users)
        self.assertTrue(users["zetta"]["default"])
        self.assertNotIn("default", users)
        ug_cfg = {"user": "zetta", "users": "default, joe"}
        (users, _groups) = self._norm(ug_cfg, distro)
        self.assertNotIn("bob", users)  # Bob is not the default now, zetta is
        self.assertIn("joe", users)
        self.assertIn("zetta", users)
        self.assertTrue(users["zetta"]["default"])
        self.assertNotIn("default", users)
        ug_cfg = {"user": "zetta", "users": ["bob", "joe"]}
        (users, _groups) = self._norm(ug_cfg, distro)
        self.assertIn("bob", users)
        self.assertIn("joe", users)
        self.assertIn("zetta", users)
        self.assertTrue(users["zetta"]["default"])
        ug_cfg = {
            "user": "zetta",
            "users": {
                "bob": True,
                "joe": True,
            },
        }
        (users, _groups) = self._norm(ug_cfg, distro)
        self.assertIn("bob", users)
        self.assertIn("joe", users)
        self.assertIn("zetta", users)
        self.assertTrue(users["zetta"]["default"])
        ug_cfg = {
            "user": "zetta",
        }
        (users, _groups) = self._norm(ug_cfg, distro)
        self.assertIn("zetta", users)
        ug_cfg = {}
        (users, groups) = self._norm(ug_cfg, distro)
        self.assertEqual({}, users)
        self.assertEqual({}, groups)

    def test_users_dict_default_additional(self):
        distro = self._make_distro("ubuntu", bcfg)
        ug_cfg = {
            "users": [{"name": "default", "blah": True}],
        }
        (users, _groups) = self._norm(ug_cfg, distro)
        self.assertIn("bob", users)
        self.assertEqual(
            ",".join(distro.get_default_user()["groups"]),
            users["bob"]["groups"],
        )
        self.assertEqual(True, users["bob"]["blah"])
        self.assertEqual(True, users["bob"]["default"])

    def test_users_dict_extract(self):
        distro = self._make_distro("ubuntu", bcfg)
        ug_cfg = {
            "users": [
                "default",
            ],
        }
        (users, _groups) = self._norm(ug_cfg, distro)
        self.assertIn("bob", users)
        (name, config) = ug_util.extract_default(users)
        self.assertEqual(name, "bob")
        expected_config = {}
        def_config = None
        try:
            def_config = distro.get_default_user()
        except NotImplementedError:
            pass
        if not def_config:
            def_config = {}
        expected_config.update(def_config)

        # Ignore these for now
        expected_config.pop("name", None)
        expected_config.pop("groups", None)
        config.pop("groups", None)
        self.assertEqual(config, expected_config)

    def test_users_dict_default(self):
        distro = self._make_distro("ubuntu", bcfg)
        ug_cfg = {
            "users": [
                "default",
            ],
        }
        (users, _groups) = self._norm(ug_cfg, distro)
        self.assertIn("bob", users)
        self.assertEqual(
            ",".join(distro.get_default_user()["groups"]),
            users["bob"]["groups"],
        )
        self.assertEqual(True, users["bob"]["default"])

    def test_users_dict_trans(self):
        distro = self._make_distro("ubuntu")
        ug_cfg = {
            "users": [
                {"name": "joe", "tr-me": True},
                {"name": "bob"},
            ],
        }
        (users, _groups) = self._norm(ug_cfg, distro)
        self.assertIn("joe", users)
        self.assertIn("bob", users)
        self.assertEqual({"tr_me": True, "default": False}, users["joe"])
        self.assertEqual({"default": False}, users["bob"])

    def test_users_dict(self):
        distro = self._make_distro("ubuntu")
        ug_cfg = {
            "users": [
                {"name": "joe"},
                {"name": "bob"},
            ],
        }
        (users, _groups) = self._norm(ug_cfg, distro)
        self.assertIn("joe", users)
        self.assertIn("bob", users)
        self.assertEqual({"default": False}, users["joe"])
        self.assertEqual({"default": False}, users["bob"])

    @mock.patch("cloudinit.subp.subp")
    def test_create_snap_user(self, mock_subp):
        mock_subp.side_effect = [
            ('{"username": "joe", "ssh-key-count": 1}\n', "")
        ]
        distro = self._make_distro("ubuntu")
        ug_cfg = {
            "users": [
                {"name": "joe", "snapuser": "joe@joe.com"},
            ],
        }
        (users, _groups) = self._norm(ug_cfg, distro)
        for (user, config) in users.items():
            print("user=%s config=%s" % (user, config))
            username = distro.create_user(user, **config)

        snapcmd = ["snap", "create-user", "--sudoer", "--json", "joe@joe.com"]
        mock_subp.assert_called_with(snapcmd, capture=True, logstring=snapcmd)
        self.assertEqual(username, "joe")

    @mock.patch("cloudinit.subp.subp")
    def test_create_snap_user_known(self, mock_subp):
        mock_subp.side_effect = [
            ('{"username": "joe", "ssh-key-count": 1}\n', "")
        ]
        distro = self._make_distro("ubuntu")
        ug_cfg = {
            "users": [
                {"name": "joe", "snapuser": "joe@joe.com", "known": True},
            ],
        }
        (users, _groups) = self._norm(ug_cfg, distro)
        for (user, config) in users.items():
            print("user=%s config=%s" % (user, config))
            username = distro.create_user(user, **config)

        snapcmd = [
            "snap",
            "create-user",
            "--sudoer",
            "--json",
            "--known",
            "joe@joe.com",
        ]
        mock_subp.assert_called_with(snapcmd, capture=True, logstring=snapcmd)
        self.assertEqual(username, "joe")

    @mock.patch("cloudinit.util.system_is_snappy")
    @mock.patch("cloudinit.util.is_group")
    @mock.patch("cloudinit.subp.subp")
    def test_add_user_on_snappy_system(
        self, mock_subp, mock_isgrp, mock_snappy
    ):
        mock_isgrp.return_value = False
        mock_subp.return_value = True
        mock_snappy.return_value = True
        distro = self._make_distro("ubuntu")
        ug_cfg = {
            "users": [
                {"name": "joe", "groups": "users", "create_groups": True},
            ],
        }
        (users, _groups) = self._norm(ug_cfg, distro)
        for (user, config) in users.items():
            print("user=%s config=%s" % (user, config))
            distro.add_user(user, **config)

        groupcmd = ["groupadd", "users", "--extrausers"]
        addcmd = ["useradd", "joe", "--extrausers", "--groups", "users", "-m"]

        mock_subp.assert_any_call(groupcmd)
        mock_subp.assert_any_call(addcmd, logstring=addcmd)


# vi: ts=4 expandtab



# This file is part of cloud-init. See LICENSE file for license information.

from tests.unittests.helpers import CiTestCase, mock
from tests.unittests.util import MockDistro


class TestManageService(CiTestCase):

    with_logs = True

    def setUp(self):
        super(TestManageService, self).setUp()
        self.dist = MockDistro()

    @mock.patch.object(MockDistro, "uses_systemd", return_value=False)
    @mock.patch("cloudinit.distros.subp.subp")
    def test_manage_service_systemctl_initcmd(self, m_subp, m_sysd):
        self.dist.init_cmd = ["systemctl"]
        self.dist.manage_service("start", "myssh")
        m_subp.assert_called_with(
            ["systemctl", "start", "myssh"], capture=True
        )

    @mock.patch.object(MockDistro, "uses_systemd", return_value=False)
    @mock.patch("cloudinit.distros.subp.subp")
    def test_manage_service_service_initcmd(self, m_subp, m_sysd):
        self.dist.init_cmd = ["service"]
        self.dist.manage_service("start", "myssh")
        m_subp.assert_called_with(["service", "myssh", "start"], capture=True)

    @mock.patch.object(MockDistro, "uses_systemd", return_value=True)
    @mock.patch("cloudinit.distros.subp.subp")
    def test_manage_service_systemctl(self, m_subp, m_sysd):
        self.dist.init_cmd = ["ignore"]
        self.dist.manage_service("start", "myssh")
        m_subp.assert_called_with(
            ["systemctl", "start", "myssh"], capture=True
        )


# vi: ts=4 sw=4 expandtab



"""Integration test for LP: #1900836.

This test mirrors the reproducing steps from the reported bug: it changes the
permissions on cloud-init.log to 600 and confirms that they remain 600 after a
reboot.
"""


def _get_log_perms(client):
    return client.execute("stat -c %a /var/log/cloud-init.log")


class TestLogPermissionsNotResetOnReboot:
    def test_permissions_unchanged(self, client):
        # Confirm that the current permissions aren't 600
        assert "644" == _get_log_perms(client)

        # Set permissions to 600 and confirm our assertion passes pre-reboot
        client.execute("chmod 600 /var/log/cloud-init.log")
        assert "600" == _get_log_perms(client)

        # Reboot
        client.restart()
        assert client.execute("cloud-init status").ok

        # Check that permissions are not reset on reboot
        assert "600" == _get_log_perms(client)



# Copyright (C) 2012 Canonical Ltd.
# Copyright (C) 2012, 2013 Hewlett-Packard Development Company, L.P.
# Copyright (C) 2012 Yahoo! Inc.
#
# Author: Scott Moser <scott.moser@canonical.com>
# Author: Juerg Haefliger <juerg.haefliger@hp.com>
# Author: Joshua Harlow <harlowja@yahoo-inc.com>
# Author: Ben Howard <ben.howard@canonical.com>
#
# This file is part of cloud-init. See LICENSE file for license information.

import abc
import os
import re
import stat
import string
import urllib.parse
from io import StringIO
from typing import Any, Mapping, Type

from cloudinit import importer
from cloudinit import log as logging
from cloudinit import net, persistence, ssh_util, subp, type_utils, util
from cloudinit.distros.parsers import hosts
from cloudinit.features import ALLOW_EC2_MIRRORS_ON_NON_AWS_INSTANCE_TYPES
from cloudinit.net import activators, eni, network_state, renderers
from cloudinit.net.network_state import parse_net_config_data

from .networking import LinuxNetworking, Networking

# Used when a cloud-config module can be run on all cloud-init distibutions.
# The value 'all' is surfaced in module documentation for distro support.
ALL_DISTROS = "all"

OSFAMILIES = {
    "alpine": ["alpine"],
    "arch": ["arch"],
    "debian": ["debian", "ubuntu"],
    "freebsd": ["freebsd"],
    "gentoo": ["gentoo"],
    "redhat": [
        "almalinux",
        "amazon",
        "centos",
        "cloudlinux",
        "eurolinux",
        "fedora",
        "miraclelinux",
        "openEuler",
        "photon",
        "rhel",
        "rocky",
        "virtuozzo",
    ],
    "suse": ["opensuse", "sles"],
}

LOG = logging.getLogger(__name__)

# This is a best guess regex, based on current EC2 AZs on 2017-12-11.
# It could break when Amazon adds new regions and new AZs.
_EC2_AZ_RE = re.compile("^[a-z][a-z]-(?:[a-z]+-)+[0-9][a-z]$")

# Default NTP Client Configurations
PREFERRED_NTP_CLIENTS = ["chrony", "systemd-timesyncd", "ntp", "ntpdate"]

# Letters/Digits/Hyphen characters, for use in domain name validation
LDH_ASCII_CHARS = string.ascii_letters + string.digits + "-"


class Distro(persistence.CloudInitPickleMixin, metaclass=abc.ABCMeta):

    usr_lib_exec = "/usr/lib"
    hosts_fn = "/etc/hosts"
    ci_sudoers_fn = "/etc/sudoers.d/90-cloud-init-users"
    hostname_conf_fn = "/etc/hostname"
    tz_zone_dir = "/usr/share/zoneinfo"
    init_cmd = ["service"]  # systemctl, service etc
    renderer_configs: Mapping[str, Mapping[str, Any]] = {}
    _preferred_ntp_clients = None
    networking_cls: Type[Networking] = LinuxNetworking
    # This is used by self.shutdown_command(), and can be overridden in
    # subclasses
    shutdown_options_map = {"halt": "-H", "poweroff": "-P", "reboot": "-r"}

    _ci_pkl_version = 1
    prefer_fqdn = False
    resolve_conf_fn = "/etc/resolv.conf"

    def __init__(self, name, cfg, paths):
        self._paths = paths
        self._cfg = cfg
        self.name = name
        self.networking: Networking = self.networking_cls()

    def _unpickle(self, ci_pkl_version: int) -> None:
        """Perform deserialization fixes for Distro."""
        if "networking" not in self.__dict__ or not self.networking.__dict__:
            # This is either a Distro pickle with no networking attribute OR
            # this is a Distro pickle with a networking attribute but from
            # before ``Networking`` had any state (meaning that
            # Networking.__setstate__ will not be called).  In either case, we
            # want to ensure that `self.networking` is freshly-instantiated:
            # either because it isn't present at all, or because it will be
            # missing expected instance state otherwise.
            self.networking = self.networking_cls()

    @abc.abstractmethod
    def install_packages(self, pkglist):
        raise NotImplementedError()

    def _write_network(self, settings):
        """Deprecated. Remove if/when arch and gentoo support renderers."""
        raise NotImplementedError(
            "Legacy function '_write_network' was called in distro '%s'.\n"
            "_write_network_config needs implementation.\n" % self.name
        )

    def _write_network_state(self, network_state):
        priority = util.get_cfg_by_path(
            self._cfg, ("network", "renderers"), None
        )

        name, render_cls = renderers.select(priority=priority)
        LOG.debug(
            "Selected renderer '%s' from priority list: %s", name, priority
        )
        renderer = render_cls(config=self.renderer_configs.get(name))
        renderer.render_network_state(network_state)

    def _find_tz_file(self, tz):
        tz_file = os.path.join(self.tz_zone_dir, str(tz))
        if not os.path.isfile(tz_file):
            raise IOError(
                "Invalid timezone %s, no file found at %s" % (tz, tz_file)
            )
        return tz_file

    def get_option(self, opt_name, default=None):
        return self._cfg.get(opt_name, default)

    def set_option(self, opt_name, value=None):
        self._cfg[opt_name] = value

    def set_hostname(self, hostname, fqdn=None):
        writeable_hostname = self._select_hostname(hostname, fqdn)
        self._write_hostname(writeable_hostname, self.hostname_conf_fn)
        self._apply_hostname(writeable_hostname)

    def uses_systemd(self):
        """Wrapper to report whether this distro uses systemd or sysvinit."""
        return uses_systemd()

    @abc.abstractmethod
    def package_command(self, command, args=None, pkgs=None):
        raise NotImplementedError()

    @abc.abstractmethod
    def update_package_sources(self):
        raise NotImplementedError()

    def get_primary_arch(self):
        arch = os.uname()[4]
        if arch in ("i386", "i486", "i586", "i686"):
            return "i386"
        return arch

    def _get_arch_package_mirror_info(self, arch=None):
        mirror_info = self.get_option("package_mirrors", [])
        if not arch:
            arch = self.get_primary_arch()
        return _get_arch_package_mirror_info(mirror_info, arch)

    def get_package_mirror_info(self, arch=None, data_source=None):
        # This resolves the package_mirrors config option
        # down to a single dict of {mirror_name: mirror_url}
        arch_info = self._get_arch_package_mirror_info(arch)
        return _get_package_mirror_info(
            data_source=data_source, mirror_info=arch_info
        )

    def apply_network(self, settings, bring_up=True):
        """Deprecated. Remove if/when arch and gentoo support renderers."""
        # this applies network where 'settings' is interfaces(5) style
        # it is obsolete compared to apply_network_config
        # Write it out

        # pylint: disable=assignment-from-no-return
        # We have implementations in arch and gentoo still
        dev_names = self._write_network(settings)
        # pylint: enable=assignment-from-no-return
        # Now try to bring them up
        if bring_up:
            return self._bring_up_interfaces(dev_names)
        return False

    def _apply_network_from_network_config(self, netconfig, bring_up=True):
        """Deprecated. Remove if/when arch and gentoo support renderers."""
        distro = self.__class__
        LOG.warning(
            "apply_network_config is not currently implemented "
            "for distribution '%s'.  Attempting to use apply_network",
            distro,
        )
        header = "\n".join(
            [
                "# Converted from network_config for distro %s" % distro,
                "# Implementation of _write_network_config is needed.",
            ]
        )
        ns = network_state.parse_net_config_data(netconfig)
        contents = eni.network_state_to_eni(
            ns, header=header, render_hwaddress=True
        )
        return self.apply_network(contents, bring_up=bring_up)

    def generate_fallback_config(self):
        return net.generate_fallback_config()

    def apply_network_config(self, netconfig, bring_up=False) -> bool:
        """Apply the network config.

        If bring_up is True, attempt to bring up the passed in devices. If
        devices is None, attempt to bring up devices returned by
        _write_network_config.

        Returns True if any devices failed to come up, otherwise False.
        """
        # This method is preferred to apply_network which only takes
        # a much less complete network config format (interfaces(5)).
        network_state = parse_net_config_data(netconfig)
        try:
            self._write_network_state(network_state)
        except NotImplementedError:
            # backwards compat until all distros have apply_network_config
            return self._apply_network_from_network_config(
                netconfig, bring_up=bring_up
            )

        # Now try to bring them up
        if bring_up:
            LOG.debug("Bringing up newly configured network interfaces")
            try:
                network_activator = activators.select_activator()
            except activators.NoActivatorException:
                LOG.warning(
                    "No network activator found, not bringing up "
                    "network interfaces"
                )
                return True
            network_activator.bring_up_all_interfaces(network_state)
        else:
            LOG.debug("Not bringing up newly configured network interfaces")
        return False

    def apply_network_config_names(self, netconfig):
        net.apply_network_config_names(netconfig)

    @abc.abstractmethod
    def apply_locale(self, locale, out_fn=None):
        raise NotImplementedError()

    @abc.abstractmethod
    def set_timezone(self, tz):
        raise NotImplementedError()

    def _get_localhost_ip(self):
        return "127.0.0.1"

    def get_locale(self):
        raise NotImplementedError()

    @abc.abstractmethod
    def _read_hostname(self, filename, default=None):
        raise NotImplementedError()

    @abc.abstractmethod
    def _write_hostname(self, hostname, filename):
        raise NotImplementedError()

    @abc.abstractmethod
    def _read_system_hostname(self):
        raise NotImplementedError()

    def _apply_hostname(self, hostname):
        # This really only sets the hostname
        # temporarily (until reboot so it should
        # not be depended on). Use the write
        # hostname functions for 'permanent' adjustments.
        LOG.debug(
            "Non-persistently setting the system hostname to %s", hostname
        )
        try:
            subp.subp(["hostname", hostname])
        except subp.ProcessExecutionError:
            util.logexc(
                LOG,
                "Failed to non-persistently adjust the system hostname to %s",
                hostname,
            )

    def _select_hostname(self, hostname, fqdn):
        # Prefer the short hostname over the long
        # fully qualified domain name
        if (
            util.get_cfg_option_bool(
                self._cfg, "prefer_fqdn_over_hostname", self.prefer_fqdn
            )
            and fqdn
        ):
            return fqdn
        if not hostname:
            return fqdn
        return hostname

    @staticmethod
    def expand_osfamily(family_list):
        distros = []
        for family in family_list:
            if family not in OSFAMILIES:
                raise ValueError(
                    "No distributions found for osfamily {}".format(family)
                )
            distros.extend(OSFAMILIES[family])
        return distros

    def update_hostname(self, hostname, fqdn, prev_hostname_fn):
        applying_hostname = hostname

        # Determine what the actual written hostname should be
        hostname = self._select_hostname(hostname, fqdn)

        # If the previous hostname file exists lets see if we
        # can get a hostname from it
        if prev_hostname_fn and os.path.exists(prev_hostname_fn):
            prev_hostname = self._read_hostname(prev_hostname_fn)
        else:
            prev_hostname = None

        # Lets get where we should write the system hostname
        # and what the system hostname is
        (sys_fn, sys_hostname) = self._read_system_hostname()
        update_files = []

        # If there is no previous hostname or it differs
        # from what we want, lets update it or create the
        # file in the first place
        if not prev_hostname or prev_hostname != hostname:
            update_files.append(prev_hostname_fn)

        # If the system hostname is different than the previous
        # one or the desired one lets update it as well
        if (not sys_hostname) or (
            sys_hostname == prev_hostname and sys_hostname != hostname
        ):
            update_files.append(sys_fn)

        # If something else has changed the hostname after we set it
        # initially, we should not overwrite those changes (we should
        # only be setting the hostname once per instance)
        if sys_hostname and prev_hostname and sys_hostname != prev_hostname:
            LOG.info(
                "%s differs from %s, assuming user maintained hostname.",
                prev_hostname_fn,
                sys_fn,
            )
            return

        # Remove duplicates (incase the previous config filename)
        # is the same as the system config filename, don't bother
        # doing it twice
        update_files = set([f for f in update_files if f])
        LOG.debug(
            "Attempting to update hostname to %s in %s files",
            hostname,
            len(update_files),
        )

        for fn in update_files:
            try:
                self._write_hostname(hostname, fn)
            except IOError:
                util.logexc(
                    LOG, "Failed to write hostname %s to %s", hostname, fn
                )

        # If the system hostname file name was provided set the
        # non-fqdn as the transient hostname.
        if sys_fn in update_files:
            self._apply_hostname(applying_hostname)

    def update_etc_hosts(self, hostname, fqdn):
        header = ""
        if os.path.exists(self.hosts_fn):
            eh = hosts.HostsConf(util.load_file(self.hosts_fn))
        else:
            eh = hosts.HostsConf("")
            header = util.make_header(base="added")
        local_ip = self._get_localhost_ip()
        prev_info = eh.get_entry(local_ip)
        need_change = False
        if not prev_info:
            eh.add_entry(local_ip, fqdn, hostname)
            need_change = True
        else:
            need_change = True
            for entry in prev_info:
                entry_fqdn = None
                entry_aliases = []
                if len(entry) >= 1:
                    entry_fqdn = entry[0]
                if len(entry) >= 2:
                    entry_aliases = entry[1:]
                if entry_fqdn is not None and entry_fqdn == fqdn:
                    if hostname in entry_aliases:
                        # Exists already, leave it be
                        need_change = False
            if need_change:
                # Doesn't exist, add that entry in...
                new_entries = list(prev_info)
                new_entries.append([fqdn, hostname])
                eh.del_entries(local_ip)
                for entry in new_entries:
                    if len(entry) == 1:
                        eh.add_entry(local_ip, entry[0])
                    elif len(entry) >= 2:
                        eh.add_entry(local_ip, *entry)
        if need_change:
            contents = StringIO()
            if header:
                contents.write("%s\n" % (header))
            contents.write("%s\n" % (eh))
            util.write_file(self.hosts_fn, contents.getvalue(), mode=0o644)

    @property
    def preferred_ntp_clients(self):
        """Allow distro to determine the preferred ntp client list"""
        if not self._preferred_ntp_clients:
            self._preferred_ntp_clients = list(PREFERRED_NTP_CLIENTS)

        return self._preferred_ntp_clients

    def _bring_up_interface(self, device_name):
        """Deprecated. Remove if/when arch and gentoo support renderers."""
        raise NotImplementedError

    def _bring_up_interfaces(self, device_names):
        """Deprecated. Remove if/when arch and gentoo support renderers."""
        am_failed = 0
        for d in device_names:
            if not self._bring_up_interface(d):
                am_failed += 1
        if am_failed == 0:
            return True
        return False

    def get_default_user(self):
        return self.get_option("default_user")

    def add_user(self, name, **kwargs):
        """
        Add a user to the system using standard GNU tools

        This should be overriden on distros where useradd is not desirable or
        not available.
        """
        # XXX need to make add_user idempotent somehow as we
        # still want to add groups or modify SSH keys on pre-existing
        # users in the image.
        if util.is_user(name):
            LOG.info("User %s already exists, skipping.", name)
            return

        if "create_groups" in kwargs:
            create_groups = kwargs.pop("create_groups")
        else:
            create_groups = True

        useradd_cmd = ["useradd", name]
        log_useradd_cmd = ["useradd", name]
        if util.system_is_snappy():
            useradd_cmd.append("--extrausers")
            log_useradd_cmd.append("--extrausers")

        # Since we are creating users, we want to carefully validate the
        # inputs. If something goes wrong, we can end up with a system
        # that nobody can login to.
        useradd_opts = {
            "gecos": "--comment",
            "homedir": "--home",
            "primary_group": "--gid",
            "uid": "--uid",
            "groups": "--groups",
            "passwd": "--password",
            "shell": "--shell",
            "expiredate": "--expiredate",
            "inactive": "--inactive",
            "selinux_user": "--selinux-user",
        }

        useradd_flags = {
            "no_user_group": "--no-user-group",
            "system": "--system",
            "no_log_init": "--no-log-init",
        }

        redact_opts = ["passwd"]

        # support kwargs having groups=[list] or groups="g1,g2"
        groups = kwargs.get("groups")
        if groups:
            if isinstance(groups, str):
                groups = groups.split(",")

            # remove any white spaces in group names, most likely
            # that came in as a string like: groups: group1, group2
            groups = [g.strip() for g in groups]

            # kwargs.items loop below wants a comma delimeted string
            # that can go right through to the command.
            kwargs["groups"] = ",".join(groups)

            primary_group = kwargs.get("primary_group")
            if primary_group:
                groups.append(primary_group)

        if create_groups and groups:
            for group in groups:
                if not util.is_group(group):
                    self.create_group(group)
                    LOG.debug("created group '%s' for user '%s'", group, name)

        # Check the values and create the command
        for key, val in sorted(kwargs.items()):

            if key in useradd_opts and val and isinstance(val, str):
                useradd_cmd.extend([useradd_opts[key], val])

                # Redact certain fields from the logs
                if key in redact_opts:
                    log_useradd_cmd.extend([useradd_opts[key], "REDACTED"])
                else:
                    log_useradd_cmd.extend([useradd_opts[key], val])

            elif key in useradd_flags and val:
                useradd_cmd.append(useradd_flags[key])
                log_useradd_cmd.append(useradd_flags[key])

        # Don't create the home directory if directed so or if the user is a
        # system user
        if kwargs.get("no_create_home") or kwargs.get("system"):
            useradd_cmd.append("-M")
            log_useradd_cmd.append("-M")
        else:
            useradd_cmd.append("-m")
            log_useradd_cmd.append("-m")

        # Run the command
        LOG.debug("Adding user %s", name)
        try:
            subp.subp(useradd_cmd, logstring=log_useradd_cmd)
        except Exception as e:
            util.logexc(LOG, "Failed to create user %s", name)
            raise e

    def add_snap_user(self, name, **kwargs):
        """
        Add a snappy user to the system using snappy tools
        """

        snapuser = kwargs.get("snapuser")
        known = kwargs.get("known", False)
        create_user_cmd = ["snap", "create-user", "--sudoer", "--json"]
        if known:
            create_user_cmd.append("--known")
        create_user_cmd.append(snapuser)

        # Run the command
        LOG.debug("Adding snap user %s", name)
        try:
            (out, err) = subp.subp(
                create_user_cmd, logstring=create_user_cmd, capture=True
            )
            LOG.debug("snap create-user returned: %s:%s", out, err)
            jobj = util.load_json(out)
            username = jobj.get("username", None)
        except Exception as e:
            util.logexc(LOG, "Failed to create snap user %s", name)
            raise e

        return username

    def create_user(self, name, **kwargs):
        """
        Creates or partially updates the ``name`` user in the system.

        This defers the actual user creation to ``self.add_user`` or
        ``self.add_snap_user``, and most of the keys in ``kwargs`` will be
        processed there if and only if the user does not already exist.

        Once the existence of the ``name`` user has been ensured, this method
        then processes these keys (for both just-created and pre-existing
        users):

        * ``plain_text_passwd``
        * ``hashed_passwd``
        * ``lock_passwd``
        * ``sudo``
        * ``ssh_authorized_keys``
        * ``ssh_redirect_user``
        """

        # Add a snap user, if requested
        if "snapuser" in kwargs:
            return self.add_snap_user(name, **kwargs)

        # Add the user
        self.add_user(name, **kwargs)

        # Set password if plain-text password provided and non-empty
        if "plain_text_passwd" in kwargs and kwargs["plain_text_passwd"]:
            self.set_passwd(name, kwargs["plain_text_passwd"])

        # Set password if hashed password is provided and non-empty
        if "hashed_passwd" in kwargs and kwargs["hashed_passwd"]:
            self.set_passwd(name, kwargs["hashed_passwd"], hashed=True)

        # Default locking down the account.  'lock_passwd' defaults to True.
        # lock account unless lock_password is False.
        if kwargs.get("lock_passwd", True):
            self.lock_passwd(name)

        # Configure sudo access
        if "sudo" in kwargs and kwargs["sudo"] is not False:
            self.write_sudo_rules(name, kwargs["sudo"])

        # Import SSH keys
        if "ssh_authorized_keys" in kwargs:
            # Try to handle this in a smart manner.
            keys = kwargs["ssh_authorized_keys"]
            if isinstance(keys, str):
                keys = [keys]
            elif isinstance(keys, dict):
                keys = list(keys.values())
            if keys is not None:
                if not isinstance(keys, (tuple, list, set)):
                    LOG.warning(
                        "Invalid type '%s' detected for"
                        " 'ssh_authorized_keys', expected list,"
                        " string, dict, or set.",
                        type(keys),
                    )
                    keys = []
                else:
                    keys = set(keys) or []
            ssh_util.setup_user_keys(set(keys), name)
        if "ssh_redirect_user" in kwargs:
            cloud_keys = kwargs.get("cloud_public_ssh_keys", [])
            if not cloud_keys:
                LOG.warning(
                    "Unable to disable SSH logins for %s given"
                    " ssh_redirect_user: %s. No cloud public-keys present.",
                    name,
                    kwargs["ssh_redirect_user"],
                )
            else:
                redirect_user = kwargs["ssh_redirect_user"]
                disable_option = ssh_util.DISABLE_USER_OPTS
                disable_option = disable_option.replace("$USER", redirect_user)
                disable_option = disable_option.replace("$DISABLE_USER", name)
                ssh_util.setup_user_keys(
                    set(cloud_keys), name, options=disable_option
                )
        return True

    def lock_passwd(self, name):
        """
        Lock the password of a user, i.e., disable password logins
        """
        # passwd must use short '-l' due to SLES11 lacking long form '--lock'
        lock_tools = (["passwd", "-l", name], ["usermod", "--lock", name])
        try:
            cmd = next(tool for tool in lock_tools if subp.which(tool[0]))
        except StopIteration as e:
            raise RuntimeError(
                "Unable to lock user account '%s'. No tools available. "
                "  Tried: %s." % (name, [c[0] for c in lock_tools])
            ) from e
        try:
            subp.subp(cmd)
        except Exception as e:
            util.logexc(LOG, "Failed to disable password for user %s", name)
            raise e

    def expire_passwd(self, user):
        try:
            subp.subp(["passwd", "--expire", user])
        except Exception as e:
            util.logexc(LOG, "Failed to set 'expire' for %s", user)
            raise e

    def set_passwd(self, user, passwd, hashed=False):
        pass_string = "%s:%s" % (user, passwd)
        cmd = ["chpasswd"]

        if hashed:
            # Need to use the short option name '-e' instead of '--encrypted'
            # (which would be more descriptive) since SLES 11 doesn't know
            # about long names.
            cmd.append("-e")

        try:
            subp.subp(cmd, pass_string, logstring="chpasswd for %s" % user)
        except Exception as e:
            util.logexc(LOG, "Failed to set password for %s", user)
            raise e

        return True

    def ensure_sudo_dir(self, path, sudo_base="/etc/sudoers"):
        # Ensure the dir is included and that
        # it actually exists as a directory
        sudoers_contents = ""
        base_exists = False
        if os.path.exists(sudo_base):
            sudoers_contents = util.load_file(sudo_base)
            base_exists = True
        found_include = False
        for line in sudoers_contents.splitlines():
            line = line.strip()
            include_match = re.search(r"^[#|@]includedir\s+(.*)$", line)
            if not include_match:
                continue
            included_dir = include_match.group(1).strip()
            if not included_dir:
                continue
            included_dir = os.path.abspath(included_dir)
            if included_dir == path:
                found_include = True
                break
        if not found_include:
            try:
                if not base_exists:
                    lines = [
                        "# See sudoers(5) for more information"
                        ' on "#include" directives:',
                        "",
                        util.make_header(base="added"),
                        "#includedir %s" % (path),
                        "",
                    ]
                    sudoers_contents = "\n".join(lines)
                    util.write_file(sudo_base, sudoers_contents, 0o440)
                else:
                    lines = [
                        "",
                        util.make_header(base="added"),
                        "#includedir %s" % (path),
                        "",
                    ]
                    sudoers_contents = "\n".join(lines)
                    util.append_file(sudo_base, sudoers_contents)
                LOG.debug("Added '#includedir %s' to %s", path, sudo_base)
            except IOError as e:
                util.logexc(LOG, "Failed to write %s", sudo_base)
                raise e
        util.ensure_dir(path, 0o750)

    def write_sudo_rules(self, user, rules, sudo_file=None):
        if not sudo_file:
            sudo_file = self.ci_sudoers_fn

        lines = [
            "",
            "# User rules for %s" % user,
        ]
        if isinstance(rules, (list, tuple)):
            for rule in rules:
                lines.append("%s %s" % (user, rule))
        elif isinstance(rules, str):
            lines.append("%s %s" % (user, rules))
        else:
            msg = "Can not create sudoers rule addition with type %r"
            raise TypeError(msg % (type_utils.obj_name(rules)))
        content = "\n".join(lines)
        content += "\n"  # trailing newline

        self.ensure_sudo_dir(os.path.dirname(sudo_file))
        if not os.path.exists(sudo_file):
            contents = [
                util.make_header(),
                content,
            ]
            try:
                util.write_file(sudo_file, "\n".join(contents), 0o440)
            except IOError as e:
                util.logexc(LOG, "Failed to write sudoers file %s", sudo_file)
                raise e
        else:
            try:
                util.append_file(sudo_file, content)
            except IOError as e:
                util.logexc(LOG, "Failed to append sudoers file %s", sudo_file)
                raise e

    def create_group(self, name, members=None):
        group_add_cmd = ["groupadd", name]
        if util.system_is_snappy():
            group_add_cmd.append("--extrausers")
        if not members:
            members = []

        # Check if group exists, and then add it doesn't
        if util.is_group(name):
            LOG.warning("Skipping creation of existing group '%s'", name)
        else:
            try:
                subp.subp(group_add_cmd)
                LOG.info("Created new group %s", name)
            except Exception:
                util.logexc(LOG, "Failed to create group %s", name)

        # Add members to the group, if so defined
        if len(members) > 0:
            for member in members:
                if not util.is_user(member):
                    LOG.warning(
                        "Unable to add group member '%s' to group '%s'"
                        "; user does not exist.",
                        member,
                        name,
                    )
                    continue

                subp.subp(["usermod", "-a", "-G", name, member])
                LOG.info("Added user '%s' to group '%s'", member, name)

    def shutdown_command(self, *, mode, delay, message):
        # called from cc_power_state_change.load_power_state
        command = ["shutdown", self.shutdown_options_map[mode]]
        try:
            if delay != "now":
                delay = "+%d" % int(delay)
        except ValueError as e:
            raise TypeError(
                "power_state[delay] must be 'now' or '+m' (minutes)."
                " found '%s'." % (delay,)
            ) from e
        args = command + [delay]
        if message:
            args.append(message)
        return args

    def manage_service(self, action, service):
        """
        Perform the requested action on a service. This handles the common
        'systemctl' and 'service' cases and may be overridden in subclasses
        as necessary.
        May raise ProcessExecutionError
        """
        init_cmd = self.init_cmd
        if self.uses_systemd() or "systemctl" in init_cmd:
            init_cmd = ["systemctl"]
            cmds = {
                "stop": ["stop", service],
                "start": ["start", service],
                "enable": ["enable", service],
                "restart": ["restart", service],
                "reload": ["reload-or-restart", service],
                "try-reload": ["reload-or-try-restart", service],
            }
        else:
            cmds = {
                "stop": [service, "stop"],
                "start": [service, "start"],
                "enable": [service, "start"],
                "restart": [service, "restart"],
                "reload": [service, "restart"],
                "try-reload": [service, "restart"],
            }
        cmd = list(init_cmd) + list(cmds[action])
        return subp.subp(cmd, capture=True)

    def set_keymap(self, layout, model, variant, options):
        if self.uses_systemd():
            subp.subp(
                [
                    "localectl",
                    "set-x11-keymap",
                    layout,
                    model,
                    variant,
                    options,
                ]
            )
        else:
            raise NotImplementedError()


def _apply_hostname_transformations_to_url(url: str, transformations: list):
    """
    Apply transformations to a URL's hostname, return transformed URL.

    This is a separate function because unwrapping and rewrapping only the
    hostname portion of a URL is complex.

    :param url:
        The URL to operate on.
    :param transformations:
        A list of ``(str) -> Optional[str]`` functions, which will be applied
        in order to the hostname portion of the URL.  If any function
        (regardless of ordering) returns None, ``url`` will be returned without
        any modification.

    :return:
        A string whose value is ``url`` with the hostname ``transformations``
        applied, or ``None`` if ``url`` is unparseable.
    """
    try:
        parts = urllib.parse.urlsplit(url)
    except ValueError:
        # If we can't even parse the URL, we shouldn't use it for anything
        return None
    new_hostname = parts.hostname
    if new_hostname is None:
        # The URL given doesn't have a hostname component, so (a) we can't
        # transform it, and (b) it won't work as a mirror; return None.
        return None

    for transformation in transformations:
        new_hostname = transformation(new_hostname)
        if new_hostname is None:
            # If a transformation returns None, that indicates we should abort
            # processing and return `url` unmodified
            return url

    new_netloc = new_hostname
    if parts.port is not None:
        new_netloc = "{}:{}".format(new_netloc, parts.port)
    return urllib.parse.urlunsplit(parts._replace(netloc=new_netloc))


def _sanitize_mirror_url(url: str):
    """
    Given a mirror URL, replace or remove any invalid URI characters.

    This performs the following actions on the URL's hostname:
      * Checks if it is an IP address, returning the URL immediately if it is
      * Converts it to its IDN form (see below for details)
      * Replaces any non-Letters/Digits/Hyphen (LDH) characters in it with
        hyphens
      * Removes any leading/trailing hyphens from each domain name label

    Before we replace any invalid domain name characters, we first need to
    ensure that any valid non-ASCII characters in the hostname will not be
    replaced, by ensuring the hostname is in its Internationalized domain name
    (IDN) representation (see RFC 5890).  This conversion has to be applied to
    the whole hostname (rather than just the substitution variables), because
    the Punycode algorithm used by IDNA transcodes each part of the hostname as
    a whole string (rather than encoding individual characters).  It cannot be
    applied to the whole URL, because (a) the Punycode algorithm expects to
    operate on domain names so doesn't output a valid URL, and (b) non-ASCII
    characters in non-hostname parts of the URL aren't encoded via Punycode.

    To put this in RFC 5890's terminology: before we remove or replace any
    characters from our domain name (which we do to ensure that each label is a
    valid LDH Label), we first ensure each label is in its A-label form.

    (Note that Python's builtin idna encoding is actually IDNA2003, not
    IDNA2008.  This changes the specifics of how some characters are encoded to
    ASCII, but doesn't affect the logic here.)

    :param url:
        The URL to operate on.

    :return:
        A sanitized version of the URL, which will have been IDNA encoded if
        necessary, or ``None`` if the generated string is not a parseable URL.
    """
    # Acceptable characters are LDH characters, plus "." to separate each label
    acceptable_chars = LDH_ASCII_CHARS + "."
    transformations = [
        # This is an IP address, not a hostname, so no need to apply the
        # transformations
        lambda hostname: None if net.is_ip_address(hostname) else hostname,
        # Encode with IDNA to get the correct characters (as `bytes`), then
        # decode with ASCII so we return a `str`
        lambda hostname: hostname.encode("idna").decode("ascii"),
        # Replace any unacceptable characters with "-"
        lambda hostname: "".join(
            c if c in acceptable_chars else "-" for c in hostname
        ),
        # Drop leading/trailing hyphens from each part of the hostname
        lambda hostname: ".".join(
            part.strip("-") for part in hostname.split(".")
        ),
    ]

    return _apply_hostname_transformations_to_url(url, transformations)


def _get_package_mirror_info(
    mirror_info, data_source=None, mirror_filter=util.search_for_mirror
):
    # given a arch specific 'mirror_info' entry (from package_mirrors)
    # search through the 'search' entries, and fallback appropriately
    # return a dict with only {name: mirror} entries.
    if not mirror_info:
        mirror_info = {}

    subst = {}
    if data_source and data_source.availability_zone:
        subst["availability_zone"] = data_source.availability_zone

        # ec2 availability zones are named cc-direction-[0-9][a-d] (us-east-1b)
        # the region is us-east-1. so region = az[0:-1]
        if _EC2_AZ_RE.match(data_source.availability_zone):
            ec2_region = data_source.availability_zone[0:-1]

            if ALLOW_EC2_MIRRORS_ON_NON_AWS_INSTANCE_TYPES:
                subst["ec2_region"] = "%s" % ec2_region
            elif data_source.platform_type == "ec2":
                subst["ec2_region"] = "%s" % ec2_region

    if data_source and data_source.region:
        subst["region"] = data_source.region

    results = {}
    for (name, mirror) in mirror_info.get("failsafe", {}).items():
        results[name] = mirror

    for (name, searchlist) in mirror_info.get("search", {}).items():
        mirrors = []
        for tmpl in searchlist:
            try:
                mirror = tmpl % subst
            except KeyError:
                continue

            mirror = _sanitize_mirror_url(mirror)
            if mirror is not None:
                mirrors.append(mirror)

        found = mirror_filter(mirrors)
        if found:
            results[name] = found

    LOG.debug("filtered distro mirror info: %s", results)

    return results


def _get_arch_package_mirror_info(package_mirrors, arch):
    # pull out the specific arch from a 'package_mirrors' config option
    default = None
    for item in package_mirrors:
        arches = item.get("arches")
        if arch in arches:
            return item
        if "default" in arches:
            default = item
    return default


def fetch(name):
    locs, looked_locs = importer.find_module(name, ["", __name__], ["Distro"])
    if not locs:
        raise ImportError(
            "No distribution found for distro %s (searched %s)"
            % (name, looked_locs)
        )
    mod = importer.import_module(locs[0])
    cls = getattr(mod, "Distro")
    return cls


def set_etc_timezone(
    tz, tz_file=None, tz_conf="/etc/timezone", tz_local="/etc/localtime"
):
    util.write_file(tz_conf, str(tz).rstrip() + "\n")
    # This ensures that the correct tz will be used for the system
    if tz_local and tz_file:
        # use a symlink if there exists a symlink or tz_local is not present
        islink = os.path.islink(tz_local)
        if islink or not os.path.exists(tz_local):
            if islink:
                util.del_file(tz_local)
            os.symlink(tz_file, tz_local)
        else:
            util.copy(tz_file, tz_local)
    return


def uses_systemd():
    try:
        res = os.lstat("/run/systemd/system")
        return stat.S_ISDIR(res.st_mode)
    except Exception:
        return False


# vi: ts=4 expandtab



# Copyright (C) 2011 Canonical Ltd.
# Copyright (C) 2012, 2013 Hewlett-Packard Development Company, L.P.
#
# Author: Scott Moser <scott.moser@canonical.com>
# Author: Juerg Haefliger <juerg.haefliger@hp.com>
#
# This file is part of cloud-init. See LICENSE file for license information.

"""
Set Hostname
------------
**Summary:** set hostname and fqdn

This module handles setting the system hostname and fqdn. If
``preserve_hostname`` is set, then the hostname will not be altered.

A hostname and fqdn can be provided by specifying a full domain name under the
``fqdn`` key. Alternatively, a hostname can be specified using the ``hostname``
key, and the fqdn of the cloud wil be used. If a fqdn specified with the
``hostname`` key, it will be handled properly, although it is better to use
the ``fqdn`` config key. If both ``fqdn`` and ``hostname`` are set,
it is distro dependent whether ``hostname`` or ``fqdn`` is used,
unless the ``prefer_fqdn_over_hostname`` option is true and fqdn is set
it will force the use of FQDN in all distros, and if false then it will
force the hostname use.

This module will run in the init-local stage before networking is configured
if the hostname is set by metadata or user data on the local system.

This will occur on datasources like nocloud and ovf where metadata and user
data are available locally. This ensures that the desired hostname is applied
before any DHCP requests are preformed on these platforms where dynamic DNS is
based on initial hostname.

**Internal name:** ``cc_set_hostname``

**Module frequency:** always

**Supported distros:** all

**Config keys**::

    preserve_hostname: <true/false>
    prefer_fqdn_over_hostname: <true/false>
    fqdn: <fqdn>
    hostname: <fqdn/hostname>
"""

import os

from cloudinit import util
from cloudinit.atomic_helper import write_json


class SetHostnameError(Exception):
    """Raised when the distro runs into an exception when setting hostname.

    This may happen if we attempt to set the hostname early in cloud-init's
    init-local timeframe as certain services may not be running yet.
    """


def handle(name, cfg, cloud, log, _args):
    if util.get_cfg_option_bool(cfg, "preserve_hostname", False):
        log.debug(
            "Configuration option 'preserve_hostname' is set,"
            " not setting the hostname in module %s",
            name,
        )
        return

    # Set prefer_fqdn_over_hostname value in distro
    hostname_fqdn = util.get_cfg_option_bool(
        cfg, "prefer_fqdn_over_hostname", None
    )
    if hostname_fqdn is not None:
        cloud.distro.set_option("prefer_fqdn_over_hostname", hostname_fqdn)

    (hostname, fqdn) = util.get_hostname_fqdn(cfg, cloud)
    # Check for previous successful invocation of set-hostname

    # set-hostname artifact file accounts for both hostname and fqdn
    # deltas. As such, it's format is different than cc_update_hostname's
    # previous-hostname file which only contains the base hostname.
    # TODO consolidate previous-hostname and set-hostname artifact files and
    # distro._read_hostname implementation so we only validate  one artifact.
    prev_fn = os.path.join(cloud.get_cpath("data"), "set-hostname")
    prev_hostname = {}
    if os.path.exists(prev_fn):
        prev_hostname = util.load_json(util.load_file(prev_fn))
    hostname_changed = hostname != prev_hostname.get(
        "hostname"
    ) or fqdn != prev_hostname.get("fqdn")
    if not hostname_changed:
        log.debug("No hostname changes. Skipping set-hostname")
        return
    log.debug("Setting the hostname to %s (%s)", fqdn, hostname)
    try:
        cloud.distro.set_hostname(hostname, fqdn)
    except Exception as e:
        msg = "Failed to set the hostname to %s (%s)" % (fqdn, hostname)
        util.logexc(log, msg)
        raise SetHostnameError("%s: %s" % (msg, e)) from e
    write_json(prev_fn, {"hostname": hostname, "fqdn": fqdn})


# vi: ts=4 expandtab



# This file is part of cloud-init. See LICENSE file for license information.
from collections import namedtuple
from unittest import mock

import pytest

from cloudinit.config.cc_install_hotplug import (
    HOTPLUG_UDEV_PATH,
    HOTPLUG_UDEV_RULES_TEMPLATE,
    handle,
)
from cloudinit.event import EventScope, EventType


@pytest.fixture()
def mocks():
    m_update_enabled = mock.patch("cloudinit.stages.update_event_enabled")
    m_write = mock.patch("cloudinit.util.write_file", autospec=True)
    m_del = mock.patch("cloudinit.util.del_file", autospec=True)
    m_subp = mock.patch("cloudinit.subp.subp")
    m_which = mock.patch("cloudinit.subp.which", return_value=None)
    m_path_exists = mock.patch("os.path.exists", return_value=False)

    yield namedtuple(
        "Mocks", "m_update_enabled m_write m_del m_subp m_which m_path_exists"
    )(
        m_update_enabled.start(),
        m_write.start(),
        m_del.start(),
        m_subp.start(),
        m_which.start(),
        m_path_exists.start(),
    )

    m_update_enabled.stop()
    m_write.stop()
    m_del.stop()
    m_subp.stop()
    m_which.stop()
    m_path_exists.stop()


class TestInstallHotplug:
    @pytest.mark.parametrize("libexec_exists", [True, False])
    def test_rules_installed_when_supported_and_enabled(
        self, mocks, libexec_exists
    ):
        mocks.m_which.return_value = "udevadm"
        mocks.m_update_enabled.return_value = True
        m_cloud = mock.MagicMock()
        m_cloud.datasource.get_supported_events.return_value = {
            EventScope.NETWORK: {EventType.HOTPLUG}
        }

        if libexec_exists:
            libexecdir = "/usr/libexec/cloud-init"
        else:
            libexecdir = "/usr/lib/cloud-init"
        with mock.patch("os.path.exists", return_value=libexec_exists):
            handle(None, {}, m_cloud, mock.Mock(), None)
            mocks.m_write.assert_called_once_with(
                filename=HOTPLUG_UDEV_PATH,
                content=HOTPLUG_UDEV_RULES_TEMPLATE.format(
                    libexecdir=libexecdir
                ),
            )
        assert mocks.m_subp.call_args_list == [
            mock.call(
                [
                    "udevadm",
                    "control",
                    "--reload-rules",
                ]
            )
        ]
        assert mocks.m_del.call_args_list == []

    def test_rules_not_installed_when_unsupported(self, mocks):
        mocks.m_update_enabled.return_value = True
        m_cloud = mock.MagicMock()
        m_cloud.datasource.get_supported_events.return_value = {}

        handle(None, {}, m_cloud, mock.Mock(), None)
        assert mocks.m_write.call_args_list == []
        assert mocks.m_del.call_args_list == []
        assert mocks.m_subp.call_args_list == []

    def test_rules_not_installed_when_disabled(self, mocks):
        mocks.m_update_enabled.return_value = False
        m_cloud = mock.MagicMock()
        m_cloud.datasource.get_supported_events.return_value = {
            EventScope.NETWORK: {EventType.HOTPLUG}
        }

        handle(None, {}, m_cloud, mock.Mock(), None)
        assert mocks.m_write.call_args_list == []
        assert mocks.m_del.call_args_list == []
        assert mocks.m_subp.call_args_list == []

    def test_rules_uninstalled_when_disabled(self, mocks):
        mocks.m_path_exists.return_value = True
        mocks.m_update_enabled.return_value = False
        m_cloud = mock.MagicMock()
        m_cloud.datasource.get_supported_events.return_value = {}

        handle(None, {}, m_cloud, mock.Mock(), None)
        mocks.m_del.assert_called_with(HOTPLUG_UDEV_PATH)
        assert mocks.m_subp.call_args_list == [
            mock.call(
                [
                    "udevadm",
                    "control",
                    "--reload-rules",
                ]
            )
        ]
        assert mocks.m_write.call_args_list == []

    def test_rules_not_installed_when_no_udevadm(self, mocks):
        mocks.m_update_enabled.return_value = True
        m_cloud = mock.MagicMock()
        m_cloud.datasource.get_supported_events.return_value = {
            EventScope.NETWORK: {EventType.HOTPLUG}
        }

        handle(None, {}, m_cloud, mock.Mock(), None)
        assert mocks.m_del.call_args_list == []
        assert mocks.m_write.call_args_list == []
        assert mocks.m_subp.call_args_list == []



# This file is part of cloud-init. See LICENSE file for license information.

import cloudinit.net.bsd
from cloudinit import log as logging
from cloudinit import subp, util

LOG = logging.getLogger(__name__)


class Renderer(cloudinit.net.bsd.BSDRenderer):
    def __init__(self, config=None):
        super(Renderer, self).__init__()

    def write_config(self):
        if self.dhcp_interfaces():
            self.set_rc_config_value("dhcpcd", "YES")
            self.set_rc_config_value(
                "dhcpcd_flags", " ".join(self.dhcp_interfaces())
            )
        for device_name, v in self.interface_configurations.items():
            if isinstance(v, dict):
                net_config = v.get("address") + " netmask " + v.get("netmask")
                mtu = v.get("mtu")
                if mtu:
                    net_config += " mtu %d" % mtu
                self.set_rc_config_value("ifconfig_" + device_name, net_config)

    def start_services(self, run=False):
        if not run:
            LOG.debug("netbsd generate postcmd disabled")
            return

        subp.subp(["service", "network", "restart"], capture=True)
        if self.dhcp_interfaces():
            subp.subp(["service", "dhcpcd", "restart"], capture=True)

    def set_route(self, network, netmask, gateway):
        if network == "0.0.0.0":
            self.set_rc_config_value("defaultroute", gateway)


def available(target=None):
    return util.is_NetBSD()



# Copyright (C) 2016 Canonical Ltd.
# Copyright (C) 2016 VMware Inc.
#
# Author: Sankar Tanguturi <stanguturi@vmware.com>
#
# This file is part of cloud-init. See LICENSE file for license information.

import logging
import os
import re
import time

from cloudinit import subp

from .guestcust_event import GuestCustEventEnum
from .guestcust_state import GuestCustStateEnum

logger = logging.getLogger(__name__)


CLOUDINIT_LOG_FILE = "/var/log/cloud-init.log"
QUERY_NICS_SUPPORTED = "queryNicsSupported"
NICS_STATUS_CONNECTED = "connected"


# This will send a RPC command to the underlying
# VMware Virtualization Platform.
def send_rpc(rpc):
    if not rpc:
        return None

    out = ""
    err = "Error sending the RPC command"

    try:
        logger.debug("Sending RPC command: %s", rpc)
        (out, err) = subp.subp(["vmware-rpctool", rpc], rcs=[0])
        # Remove the trailing newline in the output.
        if out:
            out = out.rstrip()
    except Exception as e:
        logger.debug("Failed to send RPC command")
        logger.exception(e)

    return (out, err)


# This will send the customization status to the
# underlying VMware Virtualization Platform.
def set_customization_status(custstate, custerror, errormessage=None):
    message = ""

    if errormessage:
        message = CLOUDINIT_LOG_FILE + "@" + errormessage
    else:
        message = CLOUDINIT_LOG_FILE

    rpc = "deployPkg.update.state %d %d %s" % (custstate, custerror, message)
    (out, err) = send_rpc(rpc)
    return (out, err)


def get_nics_to_enable(nicsfilepath):
    """Reads the NICS from the specified file path and returns the content

    @param nicsfilepath: Absolute file path to the NICS.txt file.
    """

    if not nicsfilepath:
        return None

    NICS_SIZE = 1024
    if not os.path.exists(nicsfilepath):
        return None

    with open(nicsfilepath, "r") as fp:
        nics = fp.read(NICS_SIZE)

    return nics


# This will send a RPC command to the underlying VMware Virtualization platform
# and enable nics.
def enable_nics(nics):
    if not nics:
        logger.warning("No Nics found")
        return

    enableNicsWaitRetries = 5
    enableNicsWaitCount = 5
    enableNicsWaitSeconds = 1

    for attempt in range(0, enableNicsWaitRetries):
        logger.debug("Trying to connect interfaces, attempt %d", attempt)
        (out, _err) = set_customization_status(
            GuestCustStateEnum.GUESTCUST_STATE_RUNNING,
            GuestCustEventEnum.GUESTCUST_EVENT_ENABLE_NICS,
            nics,
        )
        if not out:
            time.sleep(enableNicsWaitCount * enableNicsWaitSeconds)
            continue

        if out != QUERY_NICS_SUPPORTED:
            logger.warning("NICS connection status query is not supported")
            return

        for count in range(0, enableNicsWaitCount):
            (out, _err) = set_customization_status(
                GuestCustStateEnum.GUESTCUST_STATE_RUNNING,
                GuestCustEventEnum.GUESTCUST_EVENT_QUERY_NICS,
                nics,
            )
            if out and out == NICS_STATUS_CONNECTED:
                logger.info("NICS are connected on %d second", count)
                return

            time.sleep(enableNicsWaitSeconds)

    logger.warning(
        "Can't connect network interfaces after %d attempts",
        enableNicsWaitRetries,
    )


def get_tools_config(section, key, defaultVal):
    """Return the value of [section] key from VMTools configuration.

    @param section: String of section to read from VMTools config
    @returns: String value from key in [section] or defaultVal if
              [section] is not present or vmware-toolbox-cmd is
              not installed.
    """

    if not subp.which("vmware-toolbox-cmd"):
        logger.debug(
            "vmware-toolbox-cmd not installed, returning default value"
        )
        return defaultVal

    cmd = ["vmware-toolbox-cmd", "config", "get", section, key]

    try:
        (outText, _) = subp.subp(cmd)
    except subp.ProcessExecutionError as e:
        if e.exit_code == 69:
            logger.debug(
                "vmware-toolbox-cmd returned 69 (unavailable) for cmd: %s."
                " Return default value: %s",
                " ".join(cmd),
                defaultVal,
            )
        else:
            logger.error("Failed running %s[%s]", cmd, e.exit_code)
            logger.exception(e)
        return defaultVal

    retValue = defaultVal
    m = re.match(r"([^=]+)=(.*)", outText)
    if m:
        retValue = m.group(2).strip()
        logger.debug("Get tools config: [%s] %s = %s", section, key, retValue)
    else:
        logger.debug(
            "Tools config: [%s] %s is not found, return default value: %s",
            section,
            key,
            retValue,
        )

    return retValue


# Sets message to the VMX guestinfo.gc.status property to the
# underlying VMware Virtualization Platform.
def set_gc_status(config, gcMsg):
    if config and config.post_gc_status:
        rpc = "info-set guestinfo.gc.status %s" % gcMsg
        return send_rpc(rpc)
    return None


# vi: ts=4 expandtab



# This file is part of cloud-init. See LICENSE file for license information.

import shutil
import tempfile

from cloudinit import log as logging
from cloudinit import util
from cloudinit.config.cc_write_files_deferred import handle
from tests.unittests.helpers import (
    CiTestCase,
    FilesystemMockingTestCase,
    mock,
    skipUnlessJsonSchema,
)

from .test_cc_write_files import VALID_SCHEMA

LOG = logging.getLogger(__name__)


@skipUnlessJsonSchema()
@mock.patch("cloudinit.config.cc_write_files_deferred.write_files")
class TestWriteFilesDeferredSchema(CiTestCase):

    with_logs = True

    def test_schema_validation_warns_invalid_value(
        self, m_write_files_deferred
    ):
        """If 'defer' is defined, it must be of type 'bool'."""

        valid_config = {
            "write_files": [
                {**VALID_SCHEMA.get("write_files")[0], "defer": True}
            ]
        }

        invalid_config = {
            "write_files": [
                {**VALID_SCHEMA.get("write_files")[0], "defer": str("no")}
            ]
        }

        cc = self.tmp_cloud("ubuntu")
        handle("cc_write_files_deferred", valid_config, cc, LOG, [])
        self.assertNotIn(
            "Invalid cloud-config provided:", self.logs.getvalue()
        )
        handle("cc_write_files_deferred", invalid_config, cc, LOG, [])
        self.assertIn("Invalid cloud-config provided:", self.logs.getvalue())
        self.assertIn(
            "defer: 'no' is not of type 'boolean'", self.logs.getvalue()
        )


class TestWriteFilesDeferred(FilesystemMockingTestCase):

    with_logs = True

    def setUp(self):
        super(TestWriteFilesDeferred, self).setUp()
        self.tmp = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, self.tmp)

    def test_filtering_deferred_files(self):
        self.patchUtils(self.tmp)
        expected = "hello world\n"
        config = {
            "write_files": [
                {
                    "path": "/tmp/deferred.file",
                    "defer": True,
                    "content": expected,
                },
                {"path": "/tmp/not_deferred.file"},
            ]
        }
        cc = self.tmp_cloud("ubuntu")
        handle("cc_write_files_deferred", config, cc, LOG, [])
        self.assertEqual(util.load_file("/tmp/deferred.file"), expected)
        with self.assertRaises(FileNotFoundError):
            util.load_file("/tmp/not_deferred.file")


# vi: ts=4 expandtab



# This file is part of cloud-init. See LICENSE file for license information.

from cloudinit import dmi, sources
from cloudinit.sources import DataSourceEc2 as EC2

ALIYUN_PRODUCT = "Alibaba Cloud ECS"


class DataSourceAliYun(EC2.DataSourceEc2):

    dsname = "AliYun"
    metadata_urls = ["http://100.100.100.200"]

    # The minimum supported metadata_version from the ec2 metadata apis
    min_metadata_version = "2016-01-01"
    extended_metadata_versions = []

    def get_hostname(self, fqdn=False, resolve_ip=False, metadata_only=False):
        return self.metadata.get("hostname", "localhost.localdomain")

    def get_public_ssh_keys(self):
        return parse_public_keys(self.metadata.get("public-keys", {}))

    def _get_cloud_name(self):
        if _is_aliyun():
            return EC2.CloudNames.ALIYUN
        else:
            return EC2.CloudNames.NO_EC2_METADATA


def _is_aliyun():
    return dmi.read_dmi_data("system-product-name") == ALIYUN_PRODUCT


def parse_public_keys(public_keys):
    keys = []
    for _key_id, key_body in public_keys.items():
        if isinstance(key_body, str):
            keys.append(key_body.strip())
        elif isinstance(key_body, list):
            keys.extend(key_body)
        elif isinstance(key_body, dict):
            key = key_body.get("openssh-key", [])
            if isinstance(key, str):
                keys.append(key.strip())
            elif isinstance(key, list):
                keys.extend(key)
    return keys


# Used to match classes to dependencies
datasources = [
    (DataSourceAliYun, (sources.DEP_FILESYSTEM, sources.DEP_NETWORK)),
]


# Return a list of data sources that match this set of dependencies
def get_datasource_list(depends):
    return sources.list_from_depends(depends, datasources)


# vi: ts=4 expandtab



# Copyright (C) 2012 Yahoo! Inc.
#
# Author: Joshua Harlow <harlowja@yahoo-inc.com>
#
# This file is part of cloud-init. See LICENSE file for license information.

__VERSION__ = "22.1"
_PACKAGED_VERSION = "@@PACKAGED_VERSION@@"

FEATURES = [
    # supports network config version 1
    "NETWORK_CONFIG_V1",
    # supports network config version 2 (netplan)
    "NETWORK_CONFIG_V2",
]


def version_string():
    """Extract a version string from cloud-init."""
    if not _PACKAGED_VERSION.startswith("@@"):
        return _PACKAGED_VERSION
    return __VERSION__


# vi: ts=4 expandtab



"""Global conftest.py

This conftest is used for unit tests in ``cloudinit/`` and ``tests/unittests/``
as well as the integration tests in ``tests/integration_tests/``.

Any imports that are performed at the top-level here must be installed wherever
any of these tests run: that is to say, they must be listed in
``integration-requirements.txt`` and in ``test-requirements.txt``.
"""
import os
from unittest import mock

import pytest

from cloudinit import helpers, subp, util


class _FixtureUtils:
    """A namespace for fixture helper functions, used by fixture_utils.

    These helper functions are all defined as staticmethods so they are
    effectively functions; they are defined in a class only to give us a
    namespace so calling them can look like
    ``fixture_utils.fixture_util_function()`` in test code.
    """

    @staticmethod
    def closest_marker_args_or(request, marker_name: str, default):
        """Get the args for closest ``marker_name`` or return ``default``

        :param request:
            A pytest request, as passed to a fixture.
        :param marker_name:
            The name of the marker to look for
        :param default:
            The value to return if ``marker_name`` is not found.

        :return:
            The args for the closest ``marker_name`` marker, or ``default``
            if no such marker is found.
        """
        try:
            marker = request.node.get_closest_marker(marker_name)
        except AttributeError:
            # Older versions of pytest don't have the new API
            marker = request.node.get_marker(marker_name)
        if marker is not None:
            return marker.args
        return default

    @staticmethod
    def closest_marker_first_arg_or(request, marker_name: str, default):
        """Get the first arg for closest ``marker_name`` or return ``default``

        This is a convenience wrapper around closest_marker_args_or, see there
        for full details.
        """
        result = _FixtureUtils.closest_marker_args_or(
            request, marker_name, [default]
        )
        if not result:
            raise TypeError(
                "Missing expected argument to {} marker".format(marker_name)
            )
        return result[0]


@pytest.fixture(autouse=True)
def disable_subp_usage(request, fixture_utils):
    """
    Across all (pytest) tests, ensure that subp.subp is not invoked.

    Note that this can only catch invocations where the ``subp`` module is
    imported and ``subp.subp(...)`` is called.  ``from cloudinit.subp import
    subp`` imports happen before the patching here (or the CiTestCase
    monkey-patching) happens, so are left untouched.

    While ``disable_subp_usage`` unconditionally patches
    ``cloudinit.subp.subp``, any test-local patching will override this
    patching (i.e. the mock created for that patch call will replace the mock
    created by ``disable_subp_usage``), allowing tests to be written normally.
    One important exception: if ``autospec=True`` is passed to such an
    overriding patch call it will fail: autospeccing introspects the object
    being patched and as ``subp.subp`` will always be a mock when that
    autospeccing happens, the introspection fails.  (The specific error is:
    ``TypeError: name must be a str, not a MagicMock``.)

    To allow a particular test method or class to use ``subp.subp`` you can
    mark it as such::

        @pytest.mark.allow_all_subp
        def test_whoami(self):
            subp.subp(["whoami"])

    To instead allow ``subp.subp`` usage for a specific command, you can use
    the ``allow_subp_for`` mark::

        @pytest.mark.allow_subp_for("bash")
        def test_bash(self):
            subp.subp(["bash"])

    You can pass multiple commands as values; they will all be permitted::

        @pytest.mark.allow_subp_for("bash", "whoami")
        def test_several_things(self):
            subp.subp(["bash"])
            subp.subp(["whoami"])

    This fixture (roughly) mirrors the functionality of
    ``CiTestCase.allowed_subp``.  N.B. While autouse fixtures do affect
    non-pytest tests, CiTestCase's ``allowed_subp`` does take precedence (and
    we have ``TestDisableSubpUsageInTestSubclass`` to confirm that).
    """
    allow_subp_for = fixture_utils.closest_marker_args_or(
        request, "allow_subp_for", None
    )
    # Because the mark doesn't take arguments, `allow_all_subp` will be set to
    # [] if the marker is present, so explicit None checks are required
    allow_all_subp = fixture_utils.closest_marker_args_or(
        request, "allow_all_subp", None
    )

    if allow_all_subp is not None and allow_subp_for is None:
        # Only allow_all_subp specified, don't mock subp.subp
        yield
        return

    if allow_all_subp is None and allow_subp_for is None:
        # No marks, default behaviour; disallow all subp.subp usage
        def side_effect(args, *other_args, **kwargs):
            raise AssertionError("Unexpectedly used subp.subp")

    elif allow_all_subp is not None and allow_subp_for is not None:
        # Both marks, ambiguous request; raise an exception on all subp usage
        def side_effect(args, *other_args, **kwargs):
            raise AssertionError(
                "Test marked both allow_all_subp and allow_subp_for: resolve"
                " this either by modifying your test code, or by modifying"
                " disable_subp_usage to handle precedence."
            )

    else:
        # Look this up before our patch is in place, so we have access to
        # the real implementation in side_effect
        real_subp = subp.subp

        def side_effect(args, *other_args, **kwargs):
            cmd = args[0]
            if cmd not in allow_subp_for:
                raise AssertionError(
                    "Unexpectedly used subp.subp to call {} (allowed:"
                    " {})".format(cmd, ",".join(allow_subp_for))
                )
            return real_subp(args, *other_args, **kwargs)

    with mock.patch("cloudinit.subp.subp", autospec=True) as m_subp:
        m_subp.side_effect = side_effect
        yield


@pytest.fixture(scope="session")
def fixture_utils():
    """Return a namespace containing fixture utility functions.

    See :py:class:`_FixtureUtils` for further details."""
    return _FixtureUtils


@pytest.fixture
def httpretty():
    """
    Enable HTTPretty for duration of the testcase, resetting before and after.

    This will also ensure allow_net_connect is set to False, and temporarily
    unset http_proxy in os.environ if present (to work around
    https://github.com/gabrielfalcao/HTTPretty/issues/122).
    """
    import httpretty as _httpretty

    restore_proxy = os.environ.pop("http_proxy", None)
    _httpretty.HTTPretty.allow_net_connect = False
    _httpretty.reset()
    _httpretty.enable()

    yield _httpretty

    _httpretty.disable()
    _httpretty.reset()
    if restore_proxy is not None:
        os.environ["http_proxy"] = restore_proxy


@pytest.fixture
def paths(tmpdir):
    """
    Return a helpers.Paths object configured to use a tmpdir.

    (This uses the builtin tmpdir fixture.)
    """
    dirs = {
        "cloud_dir": tmpdir.mkdir("cloud_dir").strpath,
        "run_dir": tmpdir.mkdir("run_dir").strpath,
    }
    return helpers.Paths(dirs)


@pytest.fixture(autouse=True, scope="session")
def monkeypatch_system_info():
    def my_system_info():
        return {
            "platform": "invalid",
            "system": "invalid",
            "release": "invalid",
            "python": "invalid",
            "uname": ["invalid"] * 6,
            "dist": ("Distro", "-1.1", "Codename"),
            "variant": "ubuntu",
        }

    util.system_info = my_system_info



# Copyright (C) 2009-2010 Canonical Ltd.
# Copyright (C) 2012, 2013 Hewlett-Packard Development Company, L.P.
#
# Author: Scott Moser <scott.moser@canonical.com>
# Author: Juerg Haefliger <juerg.haefliger@hp.com>
#
# This file is part of cloud-init. See LICENSE file for license information.

"""
Set Passwords
-------------
**Summary:** Set user passwords and enable/disable SSH password authentication

This module consumes three top-level config keys: ``ssh_pwauth``, ``chpasswd``
and ``password``.

The ``ssh_pwauth`` config key determines whether or not sshd will be configured
to accept password authentication.  True values will enable password auth,
false values will disable password auth, and the literal string ``unchanged``
will leave it unchanged.  Setting no value will also leave the current setting
on-disk unchanged.

The ``chpasswd`` config key accepts a dictionary containing either or both of
``expire`` and ``list``.

If the ``list`` key is provided, it should contain a list of
``username:password`` pairs.  This can be either a YAML list (of strings), or a
multi-line string with one pair per line.  Each user will have the
corresponding password set.  A password can be randomly generated by specifying
``RANDOM`` or ``R`` as a user's password.  A hashed password, created by a tool
like ``mkpasswd``, can be specified; a regex
(``r'\\$(1|2a|2y|5|6)(\\$.+){2}'``) is used to determine if a password value
should be treated as a hash.

.. note::
    The users specified must already exist on the system.  Users will have been
    created by the ``cc_users_groups`` module at this point.

By default, all users on the system will have their passwords expired (meaning
that they will have to be reset the next time the user logs in).  To disable
this behaviour, set ``expire`` under ``chpasswd`` to a false value.

If a ``list`` of user/password pairs is not specified under ``chpasswd``, then
the value of the ``password`` config key will be used to set the default user's
password.

**Internal name:** ``cc_set_passwords``

**Module frequency:** per instance

**Supported distros:** all

**Config keys**::

    ssh_pwauth: <yes/no/unchanged>

    password: password1
    chpasswd:
        expire: <true/false>

    chpasswd:
        list: |
            user1:password1
            user2:RANDOM
            user3:password3
            user4:R

    ##
    # or as yaml list
    ##
    chpasswd:
        list:
            - user1:password1
            - user2:RANDOM
            - user3:password3
            - user4:R
            - user4:$6$rL..$ej...
"""

import re
from string import ascii_letters, digits

from cloudinit import log as logging
from cloudinit import subp, util
from cloudinit.distros import ug_util
from cloudinit.ssh_util import update_ssh_config

LOG = logging.getLogger(__name__)

# We are removing certain 'painful' letters/numbers
PW_SET = "".join([x for x in ascii_letters + digits if x not in "loLOI01"])


def handle_ssh_pwauth(pw_auth, distro):
    """Apply sshd PasswordAuthentication changes.

    @param pw_auth: config setting from 'pw_auth'.
                    Best given as True, False, or "unchanged".
    @param distro: an instance of the distro class for the target distribution

    @return: None"""
    cfg_name = "PasswordAuthentication"

    if util.is_true(pw_auth):
        cfg_val = "yes"
    elif util.is_false(pw_auth):
        cfg_val = "no"
    else:
        bmsg = "Leaving SSH config '%s' unchanged." % cfg_name
        if pw_auth is None or pw_auth.lower() == "unchanged":
            LOG.debug("%s ssh_pwauth=%s", bmsg, pw_auth)
        else:
            LOG.warning("%s Unrecognized value: ssh_pwauth=%s", bmsg, pw_auth)
        return

    updated = update_ssh_config({cfg_name: cfg_val})
    if not updated:
        LOG.debug("No need to restart SSH service, %s not updated.", cfg_name)
        return

    distro.manage_service("restart", distro.get_option("ssh_svcname", "ssh"))
    LOG.debug("Restarted the SSH daemon.")


def handle(_name, cfg, cloud, log, args):
    if args:
        # if run from command line, and give args, wipe the chpasswd['list']
        password = args[0]
        if "chpasswd" in cfg and "list" in cfg["chpasswd"]:
            del cfg["chpasswd"]["list"]
    else:
        password = util.get_cfg_option_str(cfg, "password", None)

    expire = True
    plist = None

    if "chpasswd" in cfg:
        chfg = cfg["chpasswd"]
        if "list" in chfg and chfg["list"]:
            if isinstance(chfg["list"], list):
                log.debug("Handling input for chpasswd as list.")
                plist = util.get_cfg_option_list(chfg, "list", plist)
            else:
                log.debug("Handling input for chpasswd as multiline string.")
                plist = util.get_cfg_option_str(chfg, "list", plist)
                if plist:
                    plist = plist.splitlines()

        expire = util.get_cfg_option_bool(chfg, "expire", expire)

    if not plist and password:
        (users, _groups) = ug_util.normalize_users_groups(cfg, cloud.distro)
        (user, _user_config) = ug_util.extract_default(users)
        if user:
            plist = ["%s:%s" % (user, password)]
        else:
            log.warning("No default or defined user to change password for.")

    errors = []
    if plist:
        plist_in = []
        hashed_plist_in = []
        hashed_users = []
        randlist = []
        users = []
        # N.B. This regex is included in the documentation (i.e. the module
        # docstring), so any changes to it should be reflected there.
        prog = re.compile(r"\$(1|2a|2y|5|6)(\$.+){2}")
        for line in plist:
            u, p = line.split(":", 1)
            if prog.match(p) is not None and ":" not in p:
                hashed_plist_in.append(line)
                hashed_users.append(u)
            else:
                # in this else branch, we potentially change the password
                # hence, a deviation from .append(line)
                if p == "R" or p == "RANDOM":
                    p = rand_user_password()
                    randlist.append("%s:%s" % (u, p))
                plist_in.append("%s:%s" % (u, p))
                users.append(u)
        ch_in = "\n".join(plist_in) + "\n"
        if users:
            try:
                log.debug("Changing password for %s:", users)
                chpasswd(cloud.distro, ch_in)
            except Exception as e:
                errors.append(e)
                util.logexc(
                    log, "Failed to set passwords with chpasswd for %s", users
                )

        hashed_ch_in = "\n".join(hashed_plist_in) + "\n"
        if hashed_users:
            try:
                log.debug("Setting hashed password for %s:", hashed_users)
                chpasswd(cloud.distro, hashed_ch_in, hashed=True)
            except Exception as e:
                errors.append(e)
                util.logexc(
                    log,
                    "Failed to set hashed passwords with chpasswd for %s",
                    hashed_users,
                )

        if len(randlist):
            blurb = (
                "Set the following 'random' passwords\n",
                "\n".join(randlist),
            )
            util.multi_log(
                "%s\n%s\n" % blurb, stderr=False, fallback_to_stdout=False
            )

        if expire:
            expired_users = []
            for u in users:
                try:
                    cloud.distro.expire_passwd(u)
                    expired_users.append(u)
                except Exception as e:
                    errors.append(e)
                    util.logexc(log, "Failed to set 'expire' for %s", u)
            if expired_users:
                log.debug("Expired passwords for: %s users", expired_users)

    handle_ssh_pwauth(cfg.get("ssh_pwauth"), cloud.distro)

    if len(errors):
        log.debug("%s errors occured, re-raising the last one", len(errors))
        raise errors[-1]


def rand_user_password(pwlen=20):
    return util.rand_str(pwlen, select_from=PW_SET)


def chpasswd(distro, plist_in, hashed=False):
    if util.is_BSD():
        for pentry in plist_in.splitlines():
            u, p = pentry.split(":")
            distro.set_passwd(u, p, hashed=hashed)
    else:
        cmd = ["chpasswd"] + (["-e"] if hashed else [])
        subp.subp(cmd, plist_in)


# vi: ts=4 expandtab



# This file is part of cloud-init. See LICENSE file for license information.
import os
from collections import namedtuple

from cloudinit import log as logging
from cloudinit import subp
from cloudinit.util import is_container, is_FreeBSD

LOG = logging.getLogger(__name__)

# Path for DMI Data
DMI_SYS_PATH = "/sys/class/dmi/id"

kdmi = namedtuple("KernelNames", ["linux", "freebsd"])
kdmi.__new__.defaults__ = (None, None)

# FreeBSD's kenv(1) and Linux /sys/class/dmi/id/* both use different names from
# dmidecode. The values are the same, and ultimately what we're interested in.
# These tools offer a "cheaper" way to access those values over dmidecode.
# This is our canonical translation table. If we add more tools on other
# platforms to find dmidecode's values, their keys need to be put in here.
DMIDECODE_TO_KERNEL = {
    "baseboard-asset-tag": kdmi("board_asset_tag", "smbios.planar.tag"),
    "baseboard-manufacturer": kdmi("board_vendor", "smbios.planar.maker"),
    "baseboard-product-name": kdmi("board_name", "smbios.planar.product"),
    "baseboard-serial-number": kdmi("board_serial", "smbios.planar.serial"),
    "baseboard-version": kdmi("board_version", "smbios.planar.version"),
    "bios-release-date": kdmi("bios_date", "smbios.bios.reldate"),
    "bios-vendor": kdmi("bios_vendor", "smbios.bios.vendor"),
    "bios-version": kdmi("bios_version", "smbios.bios.version"),
    "chassis-asset-tag": kdmi("chassis_asset_tag", "smbios.chassis.tag"),
    "chassis-manufacturer": kdmi("chassis_vendor", "smbios.chassis.maker"),
    "chassis-serial-number": kdmi("chassis_serial", "smbios.chassis.serial"),
    "chassis-version": kdmi("chassis_version", "smbios.chassis.version"),
    "system-manufacturer": kdmi("sys_vendor", "smbios.system.maker"),
    "system-product-name": kdmi("product_name", "smbios.system.product"),
    "system-serial-number": kdmi("product_serial", "smbios.system.serial"),
    "system-uuid": kdmi("product_uuid", "smbios.system.uuid"),
    "system-version": kdmi("product_version", "smbios.system.version"),
}


def _read_dmi_syspath(key):
    """
    Reads dmi data from /sys/class/dmi/id
    """
    kmap = DMIDECODE_TO_KERNEL.get(key)
    if kmap is None or kmap.linux is None:
        return None
    dmi_key_path = "{0}/{1}".format(DMI_SYS_PATH, kmap.linux)
    LOG.debug("querying dmi data %s", dmi_key_path)
    if not os.path.exists(dmi_key_path):
        LOG.debug("did not find %s", dmi_key_path)
        return None

    try:
        with open(dmi_key_path, "rb") as fp:
            key_data = fp.read()
    except PermissionError:
        LOG.debug("Could not read %s", dmi_key_path)
        return None

    # uninitialized dmi values show as all \xff and /sys appends a '\n'.
    # in that event, return empty string.
    if key_data == b"\xff" * (len(key_data) - 1) + b"\n":
        key_data = b""

    try:
        return key_data.decode("utf8").strip()
    except UnicodeDecodeError as e:
        LOG.error(
            "utf-8 decode of content (%s) in %s failed: %s",
            dmi_key_path,
            key_data,
            e,
        )

    return None


def _read_kenv(key):
    """
    Reads dmi data from FreeBSD's kenv(1)
    """
    kmap = DMIDECODE_TO_KERNEL.get(key)
    if kmap is None or kmap.freebsd is None:
        return None

    LOG.debug("querying dmi data %s", kmap.freebsd)

    try:
        cmd = ["kenv", "-q", kmap.freebsd]
        (result, _err) = subp.subp(cmd)
        result = result.strip()
        LOG.debug("kenv returned '%s' for '%s'", result, kmap.freebsd)
        return result
    except subp.ProcessExecutionError as e:
        LOG.debug("failed kenv cmd: %s\n%s", cmd, e)
        return None

    return None


def _call_dmidecode(key, dmidecode_path):
    """
    Calls out to dmidecode to get the data out. This is mostly for supporting
    OS's without /sys/class/dmi/id support.
    """
    try:
        cmd = [dmidecode_path, "--string", key]
        (result, _err) = subp.subp(cmd)
        result = result.strip()
        LOG.debug("dmidecode returned '%s' for '%s'", result, key)
        if result.replace(".", "") == "":
            return ""
        return result
    except subp.ProcessExecutionError as e:
        LOG.debug("failed dmidecode cmd: %s\n%s", cmd, e)
        return None


def read_dmi_data(key):
    """
    Wrapper for reading DMI data.

    If running in a container return None.  This is because DMI data is
    assumed to be not useful in a container as it does not represent the
    container but rather the host.

    This will do the following (returning the first that produces a
    result):
        1) Use a mapping to translate `key` from dmidecode naming to
           sysfs naming and look in /sys/class/dmi/... for a value.
        2) Use `key` as a sysfs key directly and look in /sys/class/dmi/...
        3) Fall-back to passing `key` to `dmidecode --string`.

    If all of the above fail to find a value, None will be returned.
    """

    if is_container():
        return None

    if is_FreeBSD():
        return _read_kenv(key)

    syspath_value = _read_dmi_syspath(key)
    if syspath_value is not None:
        return syspath_value

    def is_x86(arch):
        return arch == "x86_64" or (arch[0] == "i" and arch[2:] == "86")

    # running dmidecode can be problematic on some arches (LP: #1243287)
    uname_arch = os.uname()[4]
    if not (is_x86(uname_arch) or uname_arch in ("aarch64", "amd64")):
        LOG.debug("dmidata is not supported on %s", uname_arch)
        return None

    dmidecode_path = subp.which("dmidecode")
    if dmidecode_path:
        return _call_dmidecode(key, dmidecode_path)

    LOG.debug("did not find either path %s or dmidecode command", DMI_SYS_PATH)
    return None


# vi: ts=4 expandtab
